# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test functionality for the path classes from pyretis.core.path."""
import logging
import numpy as np
import unittest
from pyretis.core.path import (Path, paste_paths, check_crossing)
from pyretis.core.random_gen import RandomGenerator
from .help import (NegativeOrder, SameOrder, set_up_system,
                   PATHTEST0, PATHTEST1, PATHTEST2)

logging.disable(logging.CRITICAL)


def fill_forward_backward(pathf, pathb, npoints=20):
    """Fill in data for forward and backward paths."""
    for i in range(npoints):
        pathf.append(
            set_up_system([1.0 * i], np.zeros(3) * i, np.zeros(3) * i, i, i)
        )
        pathb.append(
            set_up_system([-1.0 * i], np.zeros(3) * i * -1.0,
                          np.zeros(3) * i * -1.0, -1.0 * i, -1.0 * i)
        )  # Pretty exotic kinetic energy.


class TestPaste(unittest.TestCase):
    """Test the paste_paths method."""

    def test_paste_paths1(self):
        """Test that we can paste paths together."""
        rgen = RandomGenerator(seed=0)
        pathf = Path(rgen, maxlen=1000)
        pathb = Path(rgen, maxlen=1000)
        fill_forward_backward(pathf, pathb, npoints=20)
        path = paste_paths(pathb, pathf, overlap=False, maxlen=None)
        self.assertEqual(path.length, 40)
        for i, phasepoint in enumerate(path.phasepoints):
            if i <= 19:
                self.assertAlmostEqual(phasepoint.order[0], i - 19.)
            else:
                self.assertAlmostEqual(phasepoint.order[0], i - 20.)
        path = paste_paths(pathb, pathf, overlap=True, maxlen=None)
        self.assertEqual(path.length, 39)
        for i, phasepoint in enumerate(path.phasepoints):
            self.assertAlmostEqual(phasepoint.order[0], i - 19.)

    def test_paste_paths2(self):
        """Test that we can paste paths together when we truncate."""
        rgen = RandomGenerator(seed=0)
        pathf = Path(rgen, maxlen=30)
        pathb = Path(rgen, maxlen=30)
        fill_forward_backward(pathf, pathb, npoints=20)
        path = paste_paths(pathb, pathf, overlap=True, maxlen=None)
        self.assertEqual(path.length, 30)
        for i, phasepoint in enumerate(path.phasepoints):
            self.assertAlmostEqual(phasepoint.order[0], i - 19.)
        pathf = Path(rgen, maxlen=32)
        pathb = Path(rgen, maxlen=31)
        fill_forward_backward(pathf, pathb, npoints=20)
        path = paste_paths(pathb, pathf, overlap=True, maxlen=None)
        self.assertEqual(path.length, 32)
        for i, phasepoint in enumerate(path.phasepoints):
            self.assertAlmostEqual(phasepoint.order[0], i - 19.)
        path = paste_paths(pathb, pathf, overlap=True, maxlen=10)
        self.assertEqual(path.length, 10)
        for i, phasepoint in enumerate(path.phasepoints):
            self.assertAlmostEqual(phasepoint.order[0], i - 19.)

    def test_paste_paths3(self):
        """Test that the backwards phasepoint survive paste-pathing."""
        rgen = RandomGenerator(seed=0)
        pathf = Path(rgen, maxlen=1000)
        pathb = Path(rgen, maxlen=1000)
        fill_forward_backward(pathf, pathb, npoints=20)
        pathb.phasepoints[0].order[0] = -11.11
        pathb.phasepoints[-1].order[0] = -22.22
        pathf.phasepoints[0].order[0] = 11.11
        pathf.phasepoints[-1].order[0] = 22.22
        path = paste_paths(pathb, pathf, overlap=True, maxlen=None)
        # The bacwkards path is the one that survives, while the
        # forwards path loses a phasepoint. Note that the backwards
        # path is being flipped in time.
        self.assertEqual(path.phasepoints[0].order[0],
                         pathb.phasepoints[-1].order[0])
        self.assertEqual(path.phasepoints[19].order[0],
                         pathb.phasepoints[0].order[0])
        self.assertEqual(path.phasepoints[20].order[0],
                         pathf.phasepoints[1].order[0])
        self.assertEqual(path.phasepoints[-1].order[0],
                         pathf.phasepoints[-1].order[0])


class TestCheckCrossing(unittest.TestCase):
    """Test the check_crossing method."""

    def test_check_crossing(self):
        """Test the check crossing method."""
        leftof, cross = check_crossing(0, 1.0, [-1.0, 0.0, 1.1], None)
        self.assertEqual(leftof, [False, False, True])
        self.assertTrue(not cross)
        leftof, cross = check_crossing(1, 1.2, [-1.0, 0.0, 1.1], leftof)
        self.assertEqual(leftof, [False, False, False])
        self.assertEqual(cross[0], (1, 2, '+'))
        leftof, cross = check_crossing(10, -2, [-1.0, 0.0, 1.1], leftof)
        self.assertEqual(leftof, [True, True, True])
        self.assertEqual(cross[0], (10, 0, '-'))
        self.assertEqual(cross[1], (10, 1, '-'))
        self.assertEqual(cross[2], (10, 2, '-'))


class TestPathProperties(unittest.TestCase):
    """Run the tests for the properties of the Path class."""

    def test_length(self):
        """Test for calculate the length of a path."""
        self.assertTrue(PATHTEST0.length == 64)
        self.assertTrue(PATHTEST2.length == 12)

    def test_minvalue(self):
        """Test for calculate the min value of a path."""
        self.assertTrue(PATHTEST0.ordermin[0] == 0)
        self.assertTrue(PATHTEST2.ordermin[0] == 4)

    def test_maxvalue(self):
        """Test for calculate the max value of a path."""
        self.assertTrue(PATHTEST0.ordermax[0] == 15)
        self.assertTrue(PATHTEST2.ordermax[0] == 6)

    def test___eq__(self):
        """Test for equal properties for paths."""
        self.assertNotEqual(PATHTEST0, PATHTEST2)
        # Compare a path with something that is not a path:
        self.assertNotEqual(PATHTEST0, 'This_is_not_a_path')
        # Compare where one path has "lost" one attribute:
        path2 = PATHTEST0.copy()
        self.assertEqual(PATHTEST0, path2)
        del path2.status
        self.assertNotEqual(PATHTEST0, path2)
        # Test when a phase point is different:
        path2 = PATHTEST0.copy()
        path2.phasepoints[-1].particles.pos = np.ones(3)
        self.assertNotEqual(PATHTEST0, path2)
        # Test with different time origin as an example of differing
        # attributes:
        path2 = PATHTEST0.copy()
        path2.time_origin = -1
        self.assertNotEqual(PATHTEST0, path2)

    def test___ne__(self):
        """Test for non equal properties for paths."""
        self.assertTrue(PATHTEST0 != PATHTEST2)


class TestPath(unittest.TestCase):
    """Run the tests for the Path class."""

    def test_path_reverse(self):
        """Test if we reverse correctly for class Path."""
        rgen = RandomGenerator(seed=0)
        path = Path(rgen)
        for _ in range(50):
            path.append(
                set_up_system([rgen.rand()[0]], np.ones(3),
                              np.ones(1), 0.0, 0.0)
            )
        path_rev = path.reverse(order_function=None)
        path_rev2 = path.reverse(order_function=SameOrder())
        self.assertTrue(path_rev == path_rev2)
        for original, rev in zip(reversed(path.phasepoints),
                                 path_rev.phasepoints):
            self.assertAlmostEqual(
                original.order[0],
                rev.order[0]
            )
            self.assertTrue(
                np.allclose(original.particles.get_vel(),
                            rev.particles.get_vel() * -1)
            )
            self.assertTrue(
                np.allclose(original.particles.get_pos(),
                            rev.particles.get_pos())
            )
        # Test for cases when the order parameter depends on the
        # velocity:
        path_rev = path.reverse(NegativeOrder())
        for original, rev in zip(reversed(path.phasepoints),
                                 path_rev.phasepoints):
            self.assertAlmostEqual(
                original.order[0],
                rev.order[0] * -1
            )
            self.assertTrue(
                np.allclose(original.particles.get_vel(),
                            rev.particles.get_vel() * -1)
            )
            self.assertTrue(
                np.allclose(original.particles.get_pos(),
                            rev.particles.get_pos())
            )

    def test_path_exceed_maxlen(self):
        """Test that we stop adding points if we exceed the path max-length."""
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=10)
        for _ in range(path.maxlen):
            add = path.append(
                set_up_system([rgen.rand()[0]], np.zeros(3),
                              np.zeros(3), 0.0, 0.0)
            )
            self.assertTrue(add)
        for _ in range(path.maxlen):
            add = path.append(
                set_up_system([rgen.rand()[0]], np.zeros(3),
                              np.zeros(3), 0.0, 0.0)
            )
            self.assertFalse(add)

    def test_empty_path_creation(self):
        """Test that empty paths are created with correct type/settings."""
        rgen = RandomGenerator(seed=0)
        maxlen = 10
        path = Path(rgen, maxlen=maxlen)
        for _ in range(maxlen + 5):
            path.append(
                set_up_system([rgen.rand()[0]], np.zeros(3),
                              np.zeros(3), 0.0, 0.0)
            )

        path2 = path.empty_path(maxlen=maxlen)

        self.assertIsInstance(path2, Path)
        self.assertEqual(path.maxlen, path2.maxlen)
        self.assertEqual(path.rgen, path2.rgen)

    def test_get_min_max(self):
        """Test the getting of min/max order parameter."""
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=100)
        all_order = []
        for i in range(20):
            order = -1.0 * i
            if i == 10:
                order = 100.
            elif i == 15:
                order = -100.
            all_order.append(order)
            path.append(
                set_up_system([order], np.zeros(3), np.zeros(3), 0.0, 0.0)
            )
        ordermin, ordermax = path.ordermin, path.ordermax
        self.assertAlmostEqual(min(all_order), ordermin[0])
        self.assertAlmostEqual(max(all_order), ordermax[0])
        self.assertAlmostEqual(15, ordermin[1])
        self.assertAlmostEqual(10, ordermax[1])

    def test_check_interfaces(self):
        """Test the check interfaces method."""
        path = Path(None, maxlen=100)
        ret = path.check_interfaces([1.0, 4.0, 5.0])
        self.assertTrue(all((i is None for i in ret)))
        for i in range(5):
            path.append(
                set_up_system([i], np.zeros(3), np.zeros(3), 0.0, 0.0)
            )
        ret = path.check_interfaces([1.0, 3.0, 5.0])
        self.assertEqual(ret[0], 'L')
        self.assertTrue(ret[1] is None)
        self.assertEqual(ret[2], 'M')
        self.assertEqual(ret[3], [True, True, False])

    def test_start_end(self):
        """Test the get start/end points method."""
        path = Path(None, maxlen=100)
        for i in range(5):
            path.append(
                set_up_system([i * -1.0], np.zeros(3), np.zeros(3), 0.0, 0.0)
            )
        end = path.get_start_point(0)
        self.assertEqual(end, 'L')
        end = path.get_end_point(1)
        self.assertEqual(end, 'L')
        end = path.get_end_point(0, 1)
        self.assertEqual(end, 'L')
        end = path.get_end_point(-10, -6)
        self.assertEqual(end, 'R')
        end = path.get_end_point(-100, -1)
        self.assertTrue(end is None)
        start = path.get_start_point(0)
        self.assertEqual(start, 'L')
        start = path.get_start_point(0, 1)
        self.assertEqual(start, 'L')
        start = path.get_start_point(-4, -3)
        self.assertEqual(start, 'R')
        start = path.get_start_point(-2, 1)
        self.assertTrue(start is None)

    def test_get_path_data(self):
        """Test the get_path_data and set_move methods."""
        path = Path(None, maxlen=100)
        path.set_move('fake')
        for i in range(5):
            path.append(
                set_up_system([i], np.zeros(3), np.zeros(3), 0.0, 0.0)
            )
        path_info = path.get_path_data('ACC', [1.0, 2.0, 3.0])
        correct = {'length': 5, 'ordermax': (4, 4), 'ordermin': (0, 0),
                   'generated': ('fake', 0, 0, 0),
                   'status': 'ACC',
                   'interface': ('L', 'M', 'R')}
        for key, val in correct.items():
            self.assertEqual(val, path_info[key])

    def test_success(self):
        """Test the success method."""
        path = Path(None, maxlen=100)
        for i in range(5):
            path.append(
                set_up_system([i], np.zeros(3), np.zeros(3), 0.0, 0.0)
            )
        self.assertTrue(path.success(3.0))
        self.assertFalse(path.success(4.0))

    def test_update_energies(self):
        """Test the update energies method."""
        path = Path(None, maxlen=100)
        for i in range(5):
            path.append(
                set_up_system([i], np.zeros(3), np.zeros(3), 1.0, 2.0)
            )
        # Test if ekin and vpot have correct length:
        ekin = [10] * path.length
        vpot = [10] * path.length
        path.update_energies(ekin, vpot)
        ekin2 = [i.particles.ekin for i in path.phasepoints]
        vpot2 = [i.particles.vpot for i in path.phasepoints]
        self.assertEqual(ekin, ekin2)
        self.assertEqual(vpot, vpot2)
        # Test for one is longer than the other:
        vpot = [11] * path.length + [12]
        path.update_energies(ekin, vpot)
        vpot2 = [i.particles.vpot for i in path.phasepoints]
        self.assertEqual(ekin, ekin2)
        self.assertEqual(vpot[:path.length], vpot2)
        # Test when energies are too short:
        ekin = [10] * (path.length - 2)
        vpot = [10] * (path.length - 1)
        path.update_energies(ekin, vpot)
        ekin2 = [i.particles.ekin for i in path.phasepoints]
        vpot2 = [i.particles.vpot for i in path.phasepoints]
        self.assertEqual(ekin + [None, None], ekin2)
        self.assertEqual(vpot + [None], vpot2)

    def test_add(self):
        """Test the __iadd__ method."""
        path = Path(None, maxlen=10)
        path2 = Path(None, maxlen=10)
        for i in range(5):
            path.append(
                set_up_system([i], np.zeros(3), np.zeros(3), 0.0, 0.0)
            )
            path2.append(
                set_up_system([i * 10], np.zeros(3), np.zeros(3), 0.0, 0.0)
            )
        path += path2
        self.assertEqual(path.length, 10)
        for i, phasepoint in enumerate(path.phasepoints):
            if i <= 4:
                self.assertEqual(phasepoint.order[0], i)
            else:
                self.assertEqual(phasepoint.order[0], (i - 5) * 10)
        # Try to add some more points (we are now > maxlen):
        path += path2
        self.assertEqual(path.length, 10)

    def test_delete(self):
        """Test the delete method."""
        path = Path(None, maxlen=10)
        for i in range(5):
            path.append(
                set_up_system([i], np.zeros(3), np.zeros(3), 0.0, 0.0)
            )
        self.assertEqual(path.length, 5)
        for i in range(path.length):
            path.delete(0)
        # Check that all points were deleted:
        self.assertEqual(path.length, 0)

    def test_sorting(self):
        """Test the sorting method."""
        data = [
            ([1, 300, 1, 2, -123], ('a', 'b'), ('a', 'b'), 1, 1),
            ([3, -4], ('e', 'f'), ('e', 'f'), 5, 4),
            ([2, 0, 0], ('c', 'd'), ('c', 'd'), 3, 2),
            ([5], ('i', 'j'), ('i', 'j'), 8, 7),
            ([4, 700000], ('g', 'h'), ('g', 'h'), 8, 7),
        ]
        path = Path(None, maxlen=10)
        for datai in data:
            path.append(
                set_up_system(datai[0], datai[1], datai[2],
                              datai[3], datai[4], internal=False)
            )
        correct_order = [1, 2, 3, 4, 5]
        correct_ekin = [1, 2, 4, 7, 7]
        correct_vpot = [1, 3, 5, 8, 8]
        # Sort path by order:
        path.sorting('order', reverse=False)
        sort = [i.order[0] for i in path.phasepoints]
        self.assertEqual(sort, correct_order)
        # Sort path by order, reversed:
        path.sorting('order', reverse=True)
        sort = [i.order[0] for i in path.phasepoints]
        self.assertEqual(sort[::-1], correct_order)
        # Sort path by kinetic energy:
        path.sorting('ekin', reverse=False)
        sort = [i.particles.ekin for i in path.phasepoints]
        self.assertEqual(sort, correct_ekin)
        # Sort path by kinetic energy, reversed:
        path.sorting('ekin', reverse=True)
        sort = [i.particles.ekin for i in path.phasepoints]
        self.assertEqual(sort[::-1], correct_ekin)
        # Sort path by potential energy:
        path.sorting('vpot', reverse=False)
        sort = [i.particles.vpot for i in path.phasepoints]
        self.assertEqual(sort, correct_vpot)
        # Sort path by potential energy, reversed:
        path.sorting('vpot', reverse=True)
        sort = [i.particles.vpot for i in path.phasepoints]
        self.assertEqual(sort[::-1], correct_vpot)
        for _ in ('gigio', 'vel'):
            with self.assertRaises(AttributeError):
                path.sorting('barbara')

        with self.assertRaises(AttributeError):
            path.sorting('gigio')


class TestExternalPaths(unittest.TestCase):
    """Test paths when we have an external particle class."""

    def test_reverse(self):
        """Test that we can reverse external paths."""
        path = Path(None)
        path.append(
            set_up_system([0.0, None], ('initial.g96', None),
                          False, None, None, internal=False)
        )
        for i in range(5):
            path.append(
                set_up_system([(i + 1) * 10, None], ('trajB.trr', i),
                              True, None, None, internal=False)
            )
        for i in range(5):
            path.append(
                set_up_system([(i + 1) * 20, None], ('trajF.trr', i),
                              False, None, None, internal=False)
            )
        rev = path.reverse(SameOrder())
        for point1, point2 in zip(rev.phasepoints,
                                  reversed(path.phasepoints)):
            self.assertNotEqual(
                point1.particles.vel_rev,
                point2.particles.vel_rev
            )
            self.assertEqual(
                point1.particles.get_pos(),
                point2.particles.get_pos()
            )
            self.assertEqual(point1.order[0], point2.order[0])

    def test_restart(self):
        """Tests the restart feature for external paths."""
        path = Path(None)
        path.append(
            set_up_system([0.0, None], ('initial.g96', None),
                          False, None, None, internal=False)
        )
        for i in range(5):
            path.append(
                set_up_system([(i + 1) * 10, None], ('trajB.trr', i),
                              True, None, None, internal=False)
            )
        for i in range(5):
            path.append(
                set_up_system([(i + 1) * 20, None], ('trajF.trr', i),
                              False, None, None, internal=False)
            )
        dic = {
            'generated': ('fake', 0, 0, 0),
            'status': 'ACC',
            'weight': 5,
        }
        for attr, value in dic.items():
            setattr(path, attr, value)
        info = path.restart_info()
        path2 = path.empty_path()
        path2.load_restart_info(info)

        for i, j in enumerate(path.rgen.get_state()['state']):
            if type(j) is np.ndarray:
                path2_l = path2.rgen.get_state()['state'][i].tolist()
                self.assertListEqual(j.tolist(), path2_l)
            else:
                self.assertEqual(j, path2.rgen.get_state()['state'][i])
        self.assertEqual(path.rgen.get_state()['seed'],
                         path2.rgen.get_state()['seed'])
        self.assertEqual(path.generated, path2.generated)
        self.assertEqual(path.maxlen, path2.maxlen)
        self.assertEqual(path.time_origin, path2.time_origin)
        self.assertEqual(path.status, path2.status)
        self.assertEqual(path.weight, path2.weight)
        for i, j in enumerate(path.phasepoints):
            self.assertEqual(j.order[0], path2.phasepoints[i].order[0])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Methods for analysing energy data from simulations.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

analyse_energies (py:func:`.analyse_energies`)
    Run the analysis for energies (kinetic, potential etc.).
"""
import numpy as np
from scipy.stats import gamma
from pyretis.analysis.analysis import analyse_data
from pyretis.core.units import CONSTANTS

np.set_printoptions(legacy='1.21')

__all__ = ['analyse_energies']


def analyse_energies(energies, settings):
    """Run the energy analysis on several energy types.

    The function will run the energy analysis on several energy types
    and collect the energies into a structure which is convenient for
    plotting the results.

    Parameters
    ----------
    energies : dict
        This dict contains the energies to analyse.
    settings : dict
        This dictionary contains settings for the analysis.

    Returns
    -------
    results : dict
        For each energy key `results[key]` contains the result from the
        energy analysis.

    See Also
    --------
    :py:func:`.analyse_data` in :py:mod:`pyretis.analysis.analysis.py`.

    """
    results = {}
    for key in energies:
        results[key] = analyse_data(energies[key], settings)
    # For the energy analysis it is also useful to add some
    # theoretical distributions:
    # Guess n particles, but overwrite it if explicitly stated.
    npart = len(settings['particles'].get('mass', []))
    npart = settings['particles'].get('npart', npart)
    alp = 0.5 * npart * settings['system']['dimensions']
    beta = settings['system'].get('beta')
    if beta is None:
        beta = (1.0 / (settings['system']['temperature'] *
                CONSTANTS['kB'][settings['system'].get('units', 'lj')]))
    scale = {'ekin': 1.0 / beta,
             'temp': settings['system']['temperature'] / alp}
    for key, value in scale.items():
        if key in results:
            dist = results[key]['distribution']
            pos = np.linspace(min(0.0, dist[1].min()), dist[1].max(), 1000)
            tdist = gamma.pdf(pos, alp, loc=0, scale=value)
            results[key]['boltzmann-dist'] = [tdist, pos, (0.0, value)]
    return results



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the CP2KEngine class."""
from io import StringIO
import logging
import os
import unittest
import tempfile
from unittest.mock import patch
import numpy as np
from pyretis.core.common import counter
from pyretis.core.system import System
from pyretis.core.path import Path
from pyretis.core.random_gen import MockRandomGenerator
from pyretis.core.particles import ParticlesExt
from pyretis.core.pathensemble import PathEnsembleExt
from pyretis.core.particlefunctions import kinetic_energy
from pyretis.engines import CP2KEngine
from pyretis.inout.common import make_dirs
from pyretis.inout.formats.xyz import (
    read_xyz_file,
    convert_snapshot,
    write_xyz_file,
)
from pyretis.orderparameter.orderparameter import PositionVelocity
from .test_helpers.test_helpers import remove_dir
logging.disable(logging.DEBUG)

HERE = os.path.abspath(os.path.dirname(__file__))


def make_test_system(conf):
    """Just make a test system with particles."""
    system = System()
    system.particles = ParticlesExt(dim=3)
    system.particles.config = conf
    return system


class CP2KEngineTest(unittest.TestCase):
    """Run the tests for the CP2KEngine."""

    def test_init(self):
        """Test that we can initiate the engine."""
        dir_name = os.path.join(HERE, 'cp2k_input')
        extra_files = ['extra_file']
        engine = CP2KEngine(cp2k='cp2k',
                            input_path=dir_name,
                            timestep=0.002,
                            subcycles=10,
                            extra_files=extra_files)
        self.assertEqual(len(engine.extra_files), 1)
        # Check that we get an error if we are missing files:
        with self.assertRaises(ValueError):
            CP2KEngine(cp2k='cp2k',
                       input_path='',
                       timestep=0.002,
                       subcycles=10,
                       extra_files=extra_files)
        # Check that non-existing extra files are not added:
        extra_files.append('a-non-existing-file-please')
        engine = CP2KEngine(cp2k='cp2k',
                            input_path=dir_name,
                            timestep=0.002,
                            subcycles=10,
                            extra_files=extra_files)
        self.assertEqual(len(engine.extra_files), 1)

    def test_single_step(self):
        """Test that the single step method work as we intend."""
        cmd = os.path.join(HERE, 'mockcp2k.py')
        dir_name = os.path.join(HERE, 'cp2k_input')
        extra_files = ['extra_file']
        engine = CP2KEngine(cp2k=cmd,
                            input_path=dir_name,
                            timestep=0.002,
                            subcycles=10,
                            extra_files=extra_files)
        rundir = os.path.join(HERE, 'generate')
        # Create the directory for running:
        make_dirs(rundir)
        engine.exe_dir = rundir
        # Create the system:
        system = make_test_system((engine.input_files['conf'], 0))
        # Run a single step:
        out = engine.step(system, 'cp2k_step')
        # Check that we have the expected files after the step:
        for i in ('extra_file', 'step.inp', 'conf.xyz', 'cp2k_step.xyz'):
            self.assertTrue(os.path.isfile(os.path.join(rundir, i)))
        self.assertAlmostEqual(system.particles.ekin, 0.9)
        self.assertAlmostEqual(system.particles.vpot, -0.9)
        # Get snapshot:
        box, xyz, vel, names = convert_snapshot(next(read_xyz_file(out)))
        for i, j in zip(names, ['H', 'H']):
            self.assertEqual(i, j)
        self.assertTrue(
            np.allclose(box,
                        np.array([1., 2., 3.]))
        )
        self.assertTrue(
            np.allclose(xyz,
                        np.array([[0.9, 1.8, 2.7], [1.9, 2.8, 3.7]]))
        )
        self.assertTrue(
            np.allclose(vel,
                        np.array([[9.9, 10.8, 11.7], [10.9, 11.8, 12.7]]))
        )
        engine.clean_up()
        remove_dir(rundir)

    def test_modify_velocities(self):
        """Test the modify velocities method."""
        cmd = os.path.join(HERE, 'mockcp2k.py')
        dir_name = os.path.join(HERE, 'cp2k_input')
        extra_files = ['extra_file']
        engine = CP2KEngine(cp2k=cmd,
                            input_path=dir_name,
                            timestep=0.002,
                            subcycles=10,
                            extra_files=extra_files)
        with tempfile.TemporaryDirectory() as tempdir:
            # Create the directory for running:
            make_dirs(tempdir)
            engine.exe_dir = tempdir
            # Create the system:
            system = make_test_system((engine.input_files['conf'], 0))
            mass = np.array([[1.], [1.]])
            system.particles.mass = mass
            system.particles.imass = np.reciprocal(mass*1.)
            system.temperature['beta'] = 1.
            pos = engine.input_files['conf']
            _, _, vel, _ = engine._read_configuration(pos)
            kin_old = kinetic_energy(vel=vel, mass=mass)[0]

            # Modify velocities:
            ensemble = {'system': system, 'rgen': MockRandomGenerator(seed=0)}
            vel_settings = {'aimless': True, 'momentum': False}
            dek, kin_new = engine.modify_velocities(ensemble, vel_settings)
            kin_new2 = kinetic_energy(vel=system.particles.vel, mass=mass)[0]
            self.assertAlmostEqual(kin_new, kin_new2)
            self.assertAlmostEqual(kin_new, 1.7070285604149718)
            self.assertAlmostEqual(dek, kin_new - kin_old)
            self.assertAlmostEqual(dek, -43.79297143958503)

            # Test aim:
            vel_settings['aimless'] = False
            vel_settings['sigma_v'] = 1
            system.particles.config = (engine.input_files['conf'], 0)
            ensemble['rgen'] = MockRandomGenerator(seed=0)
            dek, kin_new = engine.modify_velocities(ensemble, vel_settings)
            kin_new2 = kinetic_energy(vel=system.particles.vel, mass=mass)[0]
            self.assertAlmostEqual(kin_new, kin_new2)
            self.assertAlmostEqual(kin_new, 62.89068569041497)
            self.assertAlmostEqual(dek, kin_new - kin_old)
            self.assertAlmostEqual(dek, 17.39068569041497)

            # Test rescaling:
            system.particles.vpot = 2.
            vel_settings['rescale'] = 10
            system.particles.config = (engine.input_files['conf'], 0)
            dek, kin_new = engine.modify_velocities(ensemble, vel_settings)
            self.assertAlmostEqual(kin_new, 8.)
            self.assertAlmostEqual(dek, 0.)

            # Test negative rescaling:
            vel_settings['rescale'] = -2
            system.particles.config = (engine.input_files['conf'], 0)
            dek, kin_new = engine.modify_velocities(ensemble, vel_settings)
            # Velocity unchanged, so kin_new == kin_old.
            self.assertAlmostEqual(kin_new, kin_old)
            self.assertAlmostEqual(dek, 0.)

            # Test reset_momentum():
            vel_settings['momentum'] = True
            vel_settings['aimless'] = True
            vel_settings['rescale'] = False
            ensemble['rgen'] = MockRandomGenerator(seed=0)
            dek, kin_new = engine.modify_velocities(ensemble, vel_settings)
            kin_new2 = kinetic_energy(vel=system.particles.vel, mass=mass)[0]
            self.assertEqual(kin_new, kin_new2)
            self.assertEqual(dek, kin_new - kin_old)
            self.assertAlmostEqual(kin_new, 0.08269801069751077)
            self.assertAlmostEqual(dek, -45.41730198930249)

            # Test when kin_old = 0.0
            vel = np.array([[0.]*3, [0.]*3])
            zero_vel_path = os.path.join(tempdir, 'zero_vel.xyz')
            write_xyz_file(zero_vel_path, pos=vel, vel=vel)
            system.particles.config = (zero_vel_path, None)
            dek, kin_new = engine.modify_velocities(ensemble, vel_settings)
            kin_new2 = kinetic_energy(vel=system.particles.vel, mass=mass)[0]
            self.assertAlmostEqual(kin_new, kin_new2)
            self.assertAlmostEqual(kin_new, 0.23911029289543265)
            self.assertAlmostEqual(dek, float('inf'))

    def test_path_ensemble_names(self):
        """Test the path naming in path ensemble simulations."""
        cmd = os.path.join(HERE, 'mockcp2k.py')
        dir_name = os.path.join(HERE, 'cp2k_input')
        extra_files = ['extra_file']
        engine = CP2KEngine(cp2k=cmd,
                            input_path=dir_name,
                            timestep=0.002,
                            subcycles=10,
                            extra_files=extra_files)
        with tempfile.TemporaryDirectory() as tempdir:
            # Propagate in temp folder:
            engine.exe_dir = tempdir
            # Create the system:
            system = make_test_system((engine.input_files['conf'], 0))
            # Propagate:
            orderp = PositionVelocity(0, dim='x', periodic=False)
            path = Path(None, maxlen=4)
            interfaces = [0.2, 8.0, 9.0]
            p_ens = PathEnsembleExt(ensemble_number=666, interfaces=interfaces)
            ensemble = {'system': system, 'order_function': orderp,
                        'interfaces': interfaces, 'path_ensemble': p_ens}
            with patch('sys.stdout', new=StringIO()):
                success, _ = engine.propagate(path, ensemble, reverse=False)
                self.assertFalse(success)
                self.assertIn('666_', path.phasepoints[0].particles.config[0])

    def test_propagate_forward(self):
        """Test the propagate method forward in time."""
        cmd = os.path.join(HERE, 'mockcp2k.py')
        dir_name = os.path.join(HERE, 'cp2k_input')
        extra_files = ['extra_file']
        engine = CP2KEngine(cp2k=cmd,
                            input_path=dir_name,
                            timestep=0.002,
                            subcycles=10,
                            extra_files=extra_files)
        rundir = os.path.join(HERE, 'generatef')
        # Create the directory for running:
        make_dirs(rundir)
        engine.exe_dir = rundir
        # Create the system:
        system = make_test_system((engine.input_files['conf'], 0))
        # Propagate:
        orderp = PositionVelocity(0, dim='x', periodic=False)
        path = Path(None, maxlen=4)
        ensemble = {'system': system, 'order_function': orderp,
                    'interfaces': [0.2, 8.0, 9.0]}
        with patch('sys.stdout', new=StringIO()):
            success, _ = engine.propagate(path, ensemble, reverse=False)
            self.assertFalse(success)
        engine.clean_up()
        remove_dir(rundir)

    def test_propagate_backward(self):
        """Test the propagate method forward in time."""
        cmd = os.path.join(HERE, 'mockcp2k.py')
        dir_name = os.path.join(HERE, 'cp2k_input')
        extra_files = ['extra_file']
        engine = CP2KEngine(cmd,
                            input_path=dir_name,
                            timestep=0.002,
                            subcycles=10,
                            extra_files=extra_files)
        rundir = os.path.join(HERE, 'generateb')
        # Create the directory for running:
        make_dirs(rundir)
        engine.exe_dir = rundir
        # Create the system:
        system = make_test_system((engine.input_files['conf'], 0))
        # Propagate:
        orderp = PositionVelocity(0, dim='x', periodic=False)
        path = Path(None, maxlen=4)
        counter.count = -1
        ensemble = {'system': system, 'order_function': orderp,
                    'interfaces': [0.2, 0.5, 0.8]}
        with patch('sys.stdout', new=StringIO()):
            success, _ = engine.propagate(path, ensemble, reverse=True)
            self.assertTrue(success)
        # Check that initial velocities were reversed:
        infile = os.path.join(rundir, '0_conf.xyz')
        _, _, vel, _ = convert_snapshot(next(read_xyz_file(infile)))
        outfile = os.path.join(rundir, 'r_0_conf.xyz')
        _, _, rvel, _ = convert_snapshot(next(read_xyz_file(outfile)))
        self.assertTrue(np.allclose(vel, -1.0 * rvel))
        engine.clean_up()
        remove_dir(rundir)

    def test_integrate(self):
        """Test the integrate method."""
        cmd = os.path.join(HERE, 'mockcp2k.py')
        dir_name = os.path.join(HERE, 'cp2k_input')
        extra_files = ['extra_file']
        engine = CP2KEngine(cmd,
                            input_path=dir_name,
                            timestep=0.002,
                            subcycles=10,
                            extra_files=extra_files)
        rundir = os.path.join(HERE, 'cp2kintegrate')
        # Create the directory for running:
        make_dirs(rundir)
        engine.exe_dir = rundir
        # Create the system:
        system = make_test_system((engine.input_files['conf'], 0))
        # Propagate:
        orderp = PositionVelocity(0, dim='x', periodic=False)
        correct_order = [
            [0.77317158100000005, 1.0],
            [0.9, 9.9],
            [0.9, 9.9],
        ]
        ensemble = {'system': system, 'order_function': orderp}
        for i, step in enumerate(engine.integrate(ensemble, 2)):
            self.assertTrue(np.allclose(step['order'], correct_order[i]))
        # Call once more without specifying order_function
        ensemble = {'system': system}
        for step in engine.integrate(ensemble, 3):
            self.assertFalse("order" in step)
        # Call once more, after first creating a fake backup-file:
        bfile = os.path.join(rundir, 'backup.bak-1')
        with open(bfile, 'w', encoding="utf8") as output:
            output.write('Just a line')
        listoffiles = [entry.name for entry in os.scandir(rundir)]
        self.assertTrue("backup.bak-1" in listoffiles)
        for step in engine.integrate(ensemble, 1):
            self.assertFalse("order" in step)
        # Check that PyRETIS has removed the backup-file.
        listoffiles = [entry.name for entry in os.scandir(rundir)]
        self.assertFalse("backup.bak-1" in listoffiles)
        engine.clean_up()
        remove_dir(rundir)

    def test_dump_config(self):
        """Test the dump_config method."""
        cmd = os.path.join(HERE, 'mockcp2k.py')
        dir_name = os.path.join(HERE, 'cp2k_input')
        extra_files = ['extra_file']
        engine = CP2KEngine(cmd,
                            input_path=dir_name,
                            timestep=0.002,
                            subcycles=10,
                            extra_files=extra_files)
        trajdir = os.path.join(HERE, 'cp2k_fake_trajectory')
        # Create the directory for storing a fake trajectory file:
        make_dirs(trajdir)
        engine.exe_dir = trajdir
        # Set up for a trajectory file:
        trajfile = os.path.join(trajdir, 'traj.xyz')
        # Set up for a config file:
        conffile = os.path.join(trajdir, 'conf.xyz')
        lines = [
            "2",
            "# Step: 0 Box:    1.0000    1.1    1.2",
            "O         0.71    -0.361     0.332  1.1 2.2 3.6",
            "H         0.72    -0.362     0.333  1.2 2.3 3.7",
            "2",
            "# Step: 1 Box:    1.0000    1.1    1.2",
            "O         0.61    -0.331     0.232  1.5 2.2 3.7",
            "H         0.62    -0.366     0.133  2.2 3.3 4.7",
        ]
        with open(trajfile, 'w', encoding="utf8") as output:
            output.write('\n'.join(lines))
        for i in range(3):
            config = (trajfile, i)
            if i == 0:  # Config file should not exist.
                self.assertFalse(os.path.isfile(conffile))
            if i == 1:  # Config file should exist, and will be overwritten.
                self.assertTrue(os.path.isfile(conffile))
            engine.dump_config(config)
            if i == 1:
                #  Note: is it possible to check the log-file here?
                #  logger.warning('CP2K will overwrite %s', out_file)
                continue
            if i == 2:
                # Note: is it possible to check the log-file here?
                # logger.error('CP2K could not extract index %i ....
                break
            lines2 = None
            with open(conffile, 'r', encoding="utf8") as infile:
                lines2 = infile.readlines()
            len2 = len(lines2)
            for j in range(len2):
                words1 = lines[i*len2+j].split()
                words2 = lines2[j].split()
                if j == 0:
                    self.assertEqual(1, len(words2))
                    self.assertEqual(words1, words2)
                if j == 1:
                    self.assertNotEqual(words1, words2)
                    self.assertEqual(words2[1], "Box:")
                if j > 1:
                    # First item is the particle type/elements:
                    self.assertEqual(words1[0], words2[0])
                    # The rest is just numbers:
                    for k in range(1, len(words2)):
                        self.assertAlmostEqual(
                            float(words1[k]),
                            float(words2[k])
                        )
        engine.clean_up()
        remove_dir(trajdir)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the engine classes."""
import logging
import unittest
import copy
from io import StringIO
from unittest.mock import patch
import numpy as np
from pyretis.engines import (
    MDEngine,
    Langevin,
    VelocityVerlet,
    Verlet,
)
from pyretis.core import System, create_box, Particles
from pyretis.core.common import big_fat_comparer
from pyretis.core.particlefunctions import (
    calculate_thermo,
    calculate_thermo_path,
)
from pyretis.core.path import Path
from pyretis.core.random_gen import MockRandomGenerator
from pyretis.orderparameter import PositionVelocity
from pyretis.forcefield import ForceField
from pyretis.forcefield.potentials.potentials import DoubleWell
logging.disable(logging.CRITICAL)


TISMOL_POS = np.array([-1.00000000, -0.99799432, -0.98665264, -0.97520352,
                       -0.96807892, -0.96159279, -0.95014721, -0.94447852,
                       -0.94017809, -0.93804918, -0.92699119, -0.92490484,
                       -0.91348463, -0.90195902, -0.89475989, -0.88820117,
                       -0.87668488, -0.87094729, -0.86657973, -0.86438545,
                       -0.85326381, -0.85111548, -0.83963493, -0.82805057,
                       -0.82079420, -0.81417970, -0.80260905, -0.79681846,
                       -0.79239923, -0.79015456, -0.77898382, -0.77678760,
                       -0.76526037, -0.75363047, -0.74632964, -0.73967169,
                       -0.72805861, -0.72222650, -0.71776665, -0.71548223,
                       -0.70427262, -0.70203832, -0.69047381, -0.67880735,
                       -0.67147062, -0.66477741, -0.65312967, -0.64726344,
                       -0.64276997, -0.64045243, -0.62921016])

TISMOL_VEL = np.array([0.78008020, 0.78725597, 0.79204310, 0.79490675,
                       0.79431301, 0.80064533, 0.80501899, 0.80970383,
                       0.81149089, 0.81460505, 0.82094536, 0.82705648,
                       0.83080101, 0.83265977, 0.83109156, 0.83647149,
                       0.83992168, 0.84371050, 0.84461736, 0.84686145,
                       0.85235247, 0.85763502, 0.86057204, 0.86165886,
                       0.85934749, 0.86400519, 0.86676039, 0.86988007,
                       0.87013281, 0.87173251, 0.87659871, 0.88127588,
                       0.88362726, 0.88416154, 0.88132438, 0.88547579,
                       0.88774997, 0.89041257, 0.89022221, 0.89138792,
                       0.89583821, 0.90011745, 0.90208908, 0.90227395,
                       0.89911189, 0.90295628, 0.90494650, 0.90734699,
                       0.90690729, 0.90783201, 0.91205777])

TISMOL_POS_HF = np.array([-1.00000000, -0.95540084, -0.19165497, 0.78018908,
                          1.32632619, 2.28604749, 2.39185991, 2.64430617,
                          2.25348211, 2.49825935, 2.88324951, 3.30550209,
                          2.98182643, 2.76033520, 2.86756839, 2.58347298,
                          2.50423709, 2.20295643, 2.42537267, 3.05261057,
                          3.15554594, 2.44639287, 2.88716400, 3.29914672,
                          2.96754315, 3.33636578, 2.79841879, 2.84221288,
                          2.33746733, 2.54907741, 2.90952943, 3.31484556,
                          2.98322875, 2.76077697, 2.86775262, 2.58354092,
                          2.50427057, 2.20297400, 2.42538389, 3.05261681,
                          3.15554769, 2.44639327, 2.88716422, 3.29914680,
                          2.96754316, 3.33636579, 2.79841879, 2.84221288,
                          2.33746733, 2.54907741, 2.90952943])

TISMOL_VEL_HF = np.array([0.78008020, 0.04459916, 0.76596773, 0.97676712,
                          0.53799599, 0.98657113, 0.36343554, 0.55356508,
                          0.03172585, 0.48984683, 0.73416686, 0.98453450,
                          0.55129904, 0.40598753, 0.59448391, 0.26823255,
                          0.31168371, 0.05072849, 0.44876367, 0.94301707,
                          0.78008020, 0.04459916, 0.76596773, 0.97676712,
                          0.53799599, 0.98657113, 0.36343554, 0.55356508,
                          0.03172585, 0.48984683, 0.73416686, 0.98453450,
                          0.55129904, 0.40598753, 0.59448391, 0.26823255,
                          0.31168371, 0.05072849, 0.44876367, 0.94301707,
                          0.78008020, 0.04459916, 0.76596773, 0.97676712,
                          0.53799599, 0.98657113, 0.36343554, 0.55356508,
                          0.03172585, 0.48984683, 0.73416686])

TISMOL_POS_VV = np.array([-1.00000000, -0.99843984, -0.99687973, -0.99531972,
                          -0.99375986, -0.99220019, -0.99064077, -0.98908165,
                          -0.98752287, -0.98596448, -0.98440654, -0.98284908,
                          -0.98129215, -0.97973581, -0.97818009, -0.97662505,
                          -0.97507074, -0.97351719, -0.97196445, -0.97041258,
                          -0.96886161])

TISMOL_VEL_VV = np.array([0.78008020, 0.78006775, 0.78003045, 0.77996842,
                          0.77988179, 0.77977066, 0.77963517, 0.77947542,
                          0.77929154, 0.77908365, 0.77885188, 0.77859634,
                          0.77831715, 0.77801444, 0.77768833, 0.77733895,
                          0.77696642, 0.77657086, 0.77615240, 0.77571116,
                          0.77524727])


def prepare_test_system():
    """Prepare a small system we can integrate."""
    box = create_box(periodic=[False])
    system = System(temperature=1.0, units='lj', box=box)
    system.particles = Particles(system.get_dim())
    pos = np.array([-1.0])
    vel = np.array([0.78008018])
    system.add_particle(name='Ar', pos=pos, vel=vel, mass=1.0, ptype=0)
    forcefield = ForceField('Double well force field for testing')
    param = {'a': 1.0, 'b': 2.0, 'c': 0.0}
    pot = DoubleWell()
    forcefield.add_potential(pot, parameters=param)
    system.forcefield = forcefield
    return system


class EngineTest(unittest.TestCase):
    """Run the tests for the different Engines."""

    def test_restart_internal_engine_langevin(self):
        """Run test to check generic restart info."""
        system = [prepare_test_system(),
                  prepare_test_system(),
                  prepare_test_system()]

        eng = [Langevin(0.002, 0.3, rgen='mock', seed=1, high_friction=False),
               Langevin(0.002, 0.3, rgen='mock', seed=16, high_friction=False),
               Langevin(0.002, 0.3, rgen='mock', seed=1, high_friction=False)]

        for _ in range(5):
            eng[0].integration_step(system[0])
            eng[1].integration_step(system[1])
            self.assertFalse(np.allclose(system[0].particles.pos,
                                         system[1].particles.pos))

        for _ in range(16):
            eng[1].integration_step(system[1])
            eng[2].integration_step(system[2])
            self.assertFalse(np.allclose(system[1].particles.pos,
                                         system[2].particles.pos))

        sav_en = copy.deepcopy(eng[0].restart_info())
        eng[1].load_restart_info(sav_en)

        sav_sys = copy.deepcopy(system[0].restart_info())
        system[1].load_restart_info(sav_sys)

        for _ in range(11):
            eng[1].integration_step(system[1])

        self.assertTrue(np.allclose(system[1].particles.pos,
                                    system[2].particles.pos))
        self.assertTrue(np.allclose(system[1].particles.vel,
                                    system[2].particles.vel))
        self.assertFalse(np.allclose(system[0].particles.pos,
                                     system[2].particles.pos))
        self.assertFalse(np.allclose(system[0].particles.vel,
                                     system[2].particles.vel))

    def test_restart_internal_engine_velocity_verlet(self):
        """Run test to check generic restart info."""

        system = [prepare_test_system(),
                  prepare_test_system(),
                  prepare_test_system()]

        eng = [VelocityVerlet(0.02),
               VelocityVerlet(0.02),
               VelocityVerlet(0.02)]

        for _ in range(5):
            eng[0].integration_step(system[0])
            eng[1].integration_step(system[1])

        for _ in range(16):
            eng[1].integration_step(system[1])
            eng[2].integration_step(system[2])
            self.assertFalse(np.allclose(system[1].particles.pos,
                                         system[2].particles.pos))

        sav_sys = copy.deepcopy(system[0].restart_info())
        system[1].load_restart_info(sav_sys)

        for _ in range(11):
            eng[1].integration_step(system[1])

        self.assertTrue(np.allclose(system[1].particles.pos,
                                    system[2].particles.pos))
        self.assertTrue(np.allclose(system[1].particles.vel,
                                    system[2].particles.vel))
        self.assertFalse(np.allclose(system[0].particles.pos,
                                     system[2].particles.pos))
        self.assertFalse(np.allclose(system[0].particles.vel,
                                     system[2].particles.vel))

    def test_retarts_internal(self):
        """Check all internal engine restarts."""
        engs = [Verlet(timestep=0.02),
                VelocityVerlet(timestep=0.02),
                Langevin(timestep=0.02, gamma=0.1)]

        engs_base = [Verlet(timestep=1),
                     VelocityVerlet(timestep=1),
                     Langevin(timestep=1, gamma=1)]

        for engine, reference in zip(engs, engs_base):
            restart = engine.restart_info()
            reference.load_restart_info(copy.deepcopy(restart))
            if engine != reference:
                ergen = reference.rgen.get_state()
                rrgen = engine.rgen.get_state()
                # todo __eq__ method for rgen should be improved?
                self.assertTrue(big_fat_comparer(ergen, rrgen, hard=True))
                reference.rgen = engine.rgen
            self.assertTrue(engine == reference)

    def test_restart_internal_engine_verlet(self):
        """Run test to check verlet error in restart."""
        system = [prepare_test_system(),
                  prepare_test_system(),
                  prepare_test_system()]

        eng = [Verlet(timestep=0.02),
               Verlet(timestep=0.02),
               Verlet(timestep=0.02)]
        eng[0].set_initial_positions(system[0].particles)
        eng[1].set_initial_positions(system[0].particles)
        eng[2].set_initial_positions(system[0].particles)

        for _ in range(6):
            eng[0].integration_step(system[0])
            eng[1].integration_step(system[1])

        for _ in range(16):
            eng[1].integration_step(system[1])
            eng[2].integration_step(system[2])
            self.assertFalse(np.allclose(system[1].particles.pos,
                                         system[2].particles.pos))

        sav_sys = copy.deepcopy(system[0].restart_info())
        system[1].load_restart_info(sav_sys)
        eng[1].set_initial_positions(system[1].particles)

        for _ in range(10):
            eng[1].integration_step(system[1])

        self.assertFalse(np.allclose(system[1].particles.pos,
                                     system[2].particles.pos))
        self.assertFalse(np.allclose(system[1].particles.vel,
                                     system[2].particles.vel))
        self.assertFalse(np.allclose(system[0].particles.pos,
                                     system[2].particles.pos))
        self.assertFalse(np.allclose(system[0].particles.vel,
                                     system[2].particles.vel))

    def test_invert_dt(self):
        """Test that we can invert time."""
        eng = MDEngine(1, description='Mock engine for testing')
        invert = eng.invert_dt()
        self.assertFalse(invert)
        self.assertAlmostEqual(eng.timestep, -1)
        invert = eng.invert_dt()
        self.assertTrue(invert)
        self.assertAlmostEqual(eng.timestep, 1)

    def test_thermo(self):
        """Test that we can select the thermo function."""
        eng = MDEngine(1, description='Mock engine for testing')
        thermo = eng.select_thermo_function(thermo='full')
        self.assertEqual(thermo.__code__.co_code,
                         calculate_thermo.__code__.co_code)
        thermo = eng.select_thermo_function(thermo='path')
        self.assertEqual(thermo.__code__.co_code,
                         calculate_thermo_path.__code__.co_code)
        thermo = eng.select_thermo_function(thermo='gibberish')
        self.assertEqual(thermo.__code__.co_code,
                         calculate_thermo_path.__code__.co_code)
        thermo = eng.select_thermo_function(thermo=None)
        self.assertIsNone(thermo)

    def test_calculate_order(self):
        """Test that we can calculate the order parameter."""
        system = prepare_test_system()
        _box = create_box(periodic=[False], low=[0.0])
        system.box = _box
        eng = MDEngine(1, description='Mock engine for testing')
        order_function = PositionVelocity(0, dim='x', periodic=False)
        ensemble = {'system': system, 'order_function': order_function}
        order = eng.calculate_order(ensemble)
        self.assertAlmostEqual(order[0], system.particles.pos[0])
        self.assertAlmostEqual(order[1], system.particles.vel[0])
        pos = np.array([[1.0], ])
        vel = np.array([[-1.0], ])
        box = [100.0]
        order = eng.calculate_order(ensemble, xyz=pos, vel=vel, box=box)
        self.assertTrue(np.allclose(order, [1.0, -1.0]))
        self.assertTrue(np.allclose(system.box.cell, box))

    def test_modify_velocities(self):
        """Test that the engine can modify velocities."""
        system = prepare_test_system()
        eng = MDEngine(1, description='Mock engine for testing')
        rgen = MockRandomGenerator(seed=1)
        ensemble = {'system': system, 'rgen': rgen}
        eng.modify_velocities(ensemble, {'aimless': True,
                                         'momentum': False})
        self.assertAlmostEqual(system.particles.vel[0][0], rgen.rgen[1])
        deltav = np.ones_like(system.particles.vel)
        eng.modify_velocities(ensemble, {'sigma_v': deltav,
                                         'aimless': False,
                                         'momentum': False})
        self.assertAlmostEqual(system.particles.vel[0][0],
                               rgen.rgen[1] + rgen.rgen[2])
        eng.modify_velocities(ensemble, {'aimless': True,
                                         'momentum': True})
        self.assertAlmostEqual(system.particles.vel[0][0], 0.0)

        system.particles.pos = np.ones_like(system.particles.pos)
        system.particles.vpot = system.potential()
        dek, kin_new = eng.modify_velocities(ensemble,
                                             {'aimless': True,
                                              'momentum': False,
                                              'rescale': 6})
        self.assertAlmostEqual(6, kin_new + system.particles.vpot)
        kin_old = kin_new
        dek, kin_new2 = eng.modify_velocities(ensemble,
                                              {'aimless': True,
                                               'momentum': False,
                                               'rescale': -1})
        self.assertAlmostEqual(kin_old, kin_new2)
        self.assertAlmostEqual(dek, 0.0)

    def test_langevin_inertia(self):
        """Test the Langevin engine.

        In this test we compare the trajectory obtained by PyRETIS
        with a trajectory obtained independently by the TISMOL program.
        Since we are using a Langevin engine, we are using a mock
        random generator. This random generator is implemented identically
        in TISMOL and PyRETIS.
        """
        system = prepare_test_system()
        eng = Langevin(0.002, 0.3, rgen='mock', seed=1, high_friction=False)

        for i in range(51):
            self.assertTrue(np.allclose(system.particles.pos,
                                        TISMOL_POS[i]))
            self.assertTrue(np.allclose(system.particles.vel,
                                        TISMOL_VEL[i]))
            eng.integration_step(system)

    def test_langevin_negative_gamma(self):
        """Test that we fail for a negative gamma value."""
        system = prepare_test_system()
        eng = Langevin(0.002, -0.3, rgen='mock', seed=1, high_friction=False)
        with self.assertRaises(ValueError):
            eng._init_parameters(system)  # pylint: disable=protected-access

    def test_langevin_hf(self):
        """Test the high friction variant of the Langevin engine."""
        system = prepare_test_system()
        eng = Langevin(0.002, 0.3, rgen='mock', seed=1, high_friction=True)
        for i in range(51):
            self.assertTrue(np.allclose(system.particles.pos,
                                        TISMOL_POS_HF[i]))
            self.assertTrue(np.allclose(system.particles.vel,
                                        TISMOL_VEL_HF[i]))
            eng.integration_step(system)

    def test_velocityverlet(self):
        """Test the Velocity Verlet engine."""
        system = prepare_test_system()
        eng = VelocityVerlet(0.002)
        for i in range(21):
            self.assertTrue(np.allclose(system.particles.pos,
                                        TISMOL_POS_VV[i]))
            self.assertTrue(np.allclose(system.particles.vel,
                                        TISMOL_VEL_VV[i]))
            eng.integration_step(system)

    def test_velocityverlet2(self):
        """Test that the engine is callable."""
        system = prepare_test_system()
        eng = VelocityVerlet(0.002)
        eng(system)
        self.assertTrue(np.allclose(system.particles.pos, TISMOL_POS_VV[1]))
        self.assertTrue(np.allclose(system.particles.vel, TISMOL_VEL_VV[1]))

    def test_verlet(self):
        """Test the Verlet engine."""
        system = prepare_test_system()
        eng = Verlet(0.002)
        eng.set_initial_positions(system.particles)
        self.assertAlmostEqual(eng.previous_pos[0][0], -1.00156016)
        correct_pos = np.array([-0.99843984, -0.99687973, -0.99531972,
                                -0.99375986, -0.99220019])
        correct_vel = np.array([0.78008018, 0.78006773, 0.78003043,
                                0.77996841, 0.77988177])
        for i in range(5):
            eng.integration_step(system)
            self.assertAlmostEqual(system.particles.pos[0][0], correct_pos[i])
            self.assertAlmostEqual(system.particles.vel[0][0], correct_vel[i])

    def test_add_to_path(self):
        """Test that we can add to paths from the engine."""
        eng = MDEngine(1.0, 'Just testing')

        path = Path(None, maxlen=10)
        system = System()
        system.particles = Particles(system.get_dim())
        left = -1.0
        right = 8.0
        for i in range(8):
            snapshot = {'order': [1.0 * i], 'pos': np.ones(3) * i,
                        'vel': np.ones(3) * i, 'vpot': i, 'ekin': i}
            phasepoint = eng.snapshot_to_system(system, snapshot)
            status, success, stop, add = eng.add_to_path(path, phasepoint,
                                                         left, right)
            self.assertEqual(status, 'Running propagate...')
            self.assertFalse(success)
            self.assertFalse(stop)
            self.assertTrue(add)
        snapshot = {'order': [10.], 'pos': np.ones(3),
                    'vel': np.ones(3), 'vpot': 0.0, 'ekin': 0.0}
        phasepoint = eng.snapshot_to_system(system, snapshot)
        status, success, stop, add = eng.add_to_path(path, phasepoint,
                                                     left, right)
        self.assertEqual(status, 'Crossed right interface!')
        self.assertTrue(success)
        self.assertTrue(stop)
        self.assertTrue(add)
        snapshot = {'order': [9.], 'pos': np.ones(3),
                    'vel': np.ones(3), 'vpot': 0.0, 'ekin': 0.0}
        phasepoint = eng.snapshot_to_system(system, snapshot)
        status, success, stop, add = eng.add_to_path(path, phasepoint,
                                                     left, right)
        self.assertEqual(status, 'Max. path length exceeded!')
        self.assertFalse(success)
        self.assertTrue(stop)
        self.assertTrue(add)
        phasepoint = {'order': [9.], 'pos': np.ones(3),
                      'vel': np.ones(3), 'vpot': 0.0, 'ekin': 0.0}
        status, success, stop, add = eng.add_to_path(path, phasepoint,
                                                     left, right)
        self.assertEqual(status, 'Max. path length exceeded!')
        self.assertFalse(success)
        self.assertTrue(stop)
        self.assertFalse(add)

    def test_eq_and_ne(self):
        """Test that engines can be compared."""
        eng = MDEngine(1.0, 'Just testing')
        eng1 = MDEngine(1.0, 'Just Kidding')
        eng2 = VelocityVerlet(0.002)

        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(eng != eng1)
            self.assertTrue(eng != eng2)

        eng = Langevin(0.002, 0.3, rgen='mock', seed=1, high_friction=False)
        eng1 = Langevin(0.002, 0.3, rgen='rgen', seed=1, high_friction=False)
        eng2 = Langevin(0.002, 0.3)
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(eng != eng1)
            self.assertTrue(eng != eng2)

        eng = [VelocityVerlet(0.02),
               VelocityVerlet(0.02),
               VelocityVerlet(0.02),
               VelocityVerlet(0.01)]

        self.assertTrue(eng[0] == eng[1])
        self.assertTrue(eng[0] == eng[2])
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(eng[0] != eng[3])

        eng = [Langevin(0.002, 0.3, rgen='mock', seed=1, high_friction=False),
               Langevin(0.002, 0.3, rgen='mock', seed=16, high_friction=False),
               Langevin(0.002, 0.3, rgen='mock', seed=1, high_friction=False)]

        self.assertTrue(eng[0] == eng[2])
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(eng[0] != eng[1])
            self.assertTrue(eng[0] != 'gino')

        eng[0].just = 'kidding'
        self.assertTrue(eng[0] != eng[2])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""
This file defines the order parameter used for the GROMACS example.

NOTE: This file is NOT being used, it is kept as an example on how to use
      mdtraj.
      To test this file, add module = orderp.py in the Orderparameter section
      of retis-load-rc.rst file.
      Also add the import mdtraj import to the list above.

"""
import logging
import mdtraj as md
from itertools import combinations
from pyretis.orderparameter import OrderParameter
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


class Distance(OrderParameter):
    """Distance(OrderParameter).

    This class computes the distance between the O of water using mdtraj.

    Attributes
    ----------
    name : string
        A human-readable name for the order parameter.

    """

    def __init__(self, index):
        """Set up the order parameter.

        Parameters
        ----------
        index : tuple of ints
            This is the indices of the atom we will use the position of.

        """
        super().__init__(description='Water molecule distance')
        self.idx1 = index[0]
        self.idx2 = index[1]
        self.top = 'gromacs_input/conf.gro'

    def calculate(self, system):
        """Calculate the order parameter.

        Here, the order parameter is just the distance between two
        particles.

        Parameters
        ----------
        system : object like `System` from `pyretis.core.system`
            This object is used to import the file names required
            for mdtraj.

        Returns
        -------
        out : float
            The order parameter.

        """
        file_gro = system.particles.config[0]
        trj = md.load(file_gro, top=self.top)
        atom_pair = combinations([self.idx1, self.idx2], 2)
        orderp = md.compute_distances(trj, atom_pair, periodic=True)
        return orderp[0]



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This is a simple RETIS example animating the algorithm.

You can play with the interfaces, the potential parameters,
the temperature, the ratio of the different RETIS moves etc.

Have fun!
"""
import sys
import colorama
import numpy as np
import matplotlib as mpl
from matplotlib import pylab as plt
from matplotlib import animation
from matplotlib import gridspec
from pyretis.initiation import initiate_path_simulation
from pyretis.core.properties import Property
from pyretis.inout.settings import (fill_up_tis_and_retis_settings,
                                    _add_default_settings,
                                    _add_specific_default_settings)
from pyretis.setup import create_simulation
from pyretis.analysis.path_analysis import _pcross_lambda_cumulative
from pyretis.inout import print_to_screen


INTERFACES = [-0.9, -0.8, -0.7, -0.6, -0.5, -0.4, -0.3, 1.0]
PCROSS_LOG = True
# or: PCROSS_LOG = True
# Let us define the simulation:
SETTINGS = {}
# Basic settings for the simulation:
SETTINGS['simulation'] = {
    'task': 'retis',
    'steps': 150,
    'interfaces': INTERFACES
}
# Basic settings for the simulation box:
SETTINGS['box'] = {'cell': [3]}

# Basic settings for the system:
SETTINGS['system'] = {'units': 'reduced',
                      'temperature': 0.09,
                      'dimensions': 1}
# Basic settings for the Langevin integrator:
SETTINGS['engine'] = {
    'class': 'Langevin',
    'gamma': 0.3,
    'high_friction': False,
    'seed': 0,
    'timestep': 0.002,
}
# Particles:
SETTINGS['particles'] = {'mass': {'Ar': 1.0},
                         'name': ['Ar'],
                         'ptype': [0],
                         'position': {'generate': '1d'}}

# Potential parameters:
# The potential is: `V_\text{pot} = a x^4 - b (x - c)^2`
SETTINGS['potential'] = [
    {'a': 1.0, 'b': 2.0, 'c': 0.0, 'class': 'DoubleWell'}
]
# Settings for the order parameter:
SETTINGS['orderparameter'] = {
    'class': 'PositionVelocity',
    'dim': 'x',
    'index': 0,
    'periodic': False,
}
# TIS specific settings:
SETTINGS['tis'] = {
    'freq': 0.5,
    'maxlength': 20000,
    'aimless': True,
    'allowmaxlength': False,
    'sigma_v': -1,
    'seed': 0,
    'zero_momentum': False,
    'rescale_energy': False,
}
SETTINGS['initial-path'] = {'method': 'kick'}
# RETIS specific settings:
SETTINGS['retis'] = {
    'swapfreq': 0.5,
    'relative_shoots': None,
    'nullmoves': True,
    'swapsimul': True,
}

# For convenience:
TIMESTEP = SETTINGS['engine']['timestep']
ANALYSIS = {'ngrid': 100, 'nblock': 5}

_add_default_settings(SETTINGS)
_add_specific_default_settings(SETTINGS)
fill_up_tis_and_retis_settings(SETTINGS)

# Set up for plotting:
mpl.rc('font', size=14, family='serif')
mpl.rc('lines', color='#262626')
mpl.rc('savefig', directory=None)
NINT = len(INTERFACES)
CMAP = plt.get_cmap('Set1')
COLORS = [CMAP(float(i)/float(NINT)) for i in range(NINT)]
TXTCOLOR = {
    'SW': '#006BA4',
    'NU': '#FF800E',
    'TR': '#ABABAB',
    'SH': '#595959',
    'IN': '#808080',
}
FTOT = 0


def get_path(path):
    """Get points on a trajectory, useful for plotting.

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        The path/trajectory we are collecting points from.

    Returns
    -------
    out[0] : numpy.array
        The order parameter(s) as a function of time.
    out[1] : numpy.array
        The positions as a function of time.
    out[2] : numpy.array
        The velocities as a function of time.

    """
    order = []
    pos = []
    vel = []
    leng = path.length
    if leng <= 50:
        freq = 1
    elif 50 < leng <= 600:
        freq = 6
    elif 600 < leng < 2000:
        freq = 10
    else:
        freq = 20
    for i, point in enumerate(path.phasepoints):
        if i == 0 or i == leng-1 or i % freq == 0:
            order.append(point.order)
            pos.append(point.particles.pos[0])
            vel.append(point.particles.vel[0])
    return np.array(order), np.array(pos), np.array(vel)


def step_txt(ensembles, retis_result, prun):
    """Return some text information about the RETIS step.

    Parameters
    ----------
    ensembles : list of dicts, it contains:
       *  path_ensemble: objects like :py:class:`.PathEnsemble`
               The different path ensembles we are simulating.
    retis_result : list of lists
        The results from a RETIS simulation step.
    prun : list of floats
        The running average for the ensemble crossing probabilities.

    Returns
    -------
    out : list of strings
        Each string contains information about the move performed
        in an ensemble.

    """
    txt = []
    force = 0  # counter for force evaluations
    # The force counter will just count the number of
    # force evaluations for generating a path, it will not include
    # the force evaluation that might be performed prior to propagation,
    # i.e. the initial force evaluation since this can be stored in memory,
    # along the path -> i.e. we can avoid it.
    for ensemble in ensembles:
        path_ensemble = ensemble['path_ensemble']
        name = path_ensemble.ensemble_name
        idx = path_ensemble.ensemble_number
        name_of_move = retis_result[f'move-{idx}']
        accepted = retis_result[f'accept-{idx}']
        line = []
        if name_of_move == 'swap':
            idx2 = retis_result[f'all-{idx}']['swap-with']
            name2 = ensembles[idx2]['path_ensemble'].ensemble_name
            move = f'{name_of_move} {name2},'
            if idx == 0 or (idx == 1 and idx2 == 0):
                # Evaluate forces when swapping [0^-] <-> [0^+]:
                force += path_ensemble.paths[-1]['length'] - 2
        elif name_of_move == 'tis':
            trial_path = retis_result[f'path-{idx}']
            tis_move = trial_path.generated[0]
            move = f'{name_of_move} ({tis_move}),'
            if tis_move == 'sh':
                force += path_ensemble.paths[-1]['length'] - 1
        else:
            move = f'{name_of_move},'
        line.append(f'{name}: {move:11s}')
        line.append(f'{accepted},')
        if idx > 0:
            line.append(f'p = {prun[idx]:<8.6g}')
        txt.append(' '.join(line))
    return txt, force


def probability_path_ensemble(path_ensemble, step, prun, orderp):
    """Update running estimates of probabilities for an ensemble.

    Parameters
    ----------
    path_ensemble : objects like :py:class:`.PathEnsemble`
        The ensemble to analyse.
    step : int
        The current simulation step.
    prun : float
        Current running estimate for the crossing probability
        for this ensemble.
    orderp : numpy.array
        The maximum order parameters for all accepted paths.

    Returns
    -------
    prun : float
        Updated running average for the crossing probability.
    ordernew : numpy.array
        Updated list of all order parameters seen.
    lamb : numpy.array
        Coordinates used for obtaining the crossing probability
        as a function of the order parameter.
    pcross : numpy.array
        The crossing probability as a function of the order
        parameter.

    """
    ordermax = path_ensemble.last_path.ordermax[0]
    success = 1 if ordermax > path_ensemble.detect else 0
    if prun is None:
        prun = success
    else:
        npath = step + 1
        prun = float(success + prun * (npath - 1)) / float(npath)
    if orderp is None:
        ordernew = np.array([ordermax])
    else:
        ordernew = np.append(orderp, ordermax)
    ordermax = min(max(ordernew), max(path_ensemble.interfaces))
    ordermin = path_ensemble.interfaces[1]
    pcross, lamb = _pcross_lambda_cumulative(ordernew, ordermin, ordermax,
                                             ANALYSIS['ngrid'])
    return prun, ordernew, lamb, pcross


def simple_block(data, nblock):
    """Block error analysis for running average."""
    length = len(data)
    if length < nblock:
        return float('inf')
    skip, _ = divmod(len(data), nblock)
    run_avg = []
    block_avg = []
    for i in range(nblock):
        idx = (i + 1) * skip - 1
        run_avg.append(data[idx])
        if i == 0:
            block_avg.append(run_avg[i])
        else:
            block_avg.append((i + 1)*run_avg[i] - i*run_avg[i-1])
    var = sum([(x - data[-1])**2 for x in block_avg]) / (nblock - 1)
    err = np.sqrt(var / nblock)
    return err


def analyse_path_ensembles(ensembles, step, variables):
    """Calculate some properties for the path ensembles.

    Here we obtain the crossing probabilities and the initial flux.

    Parameters
    ----------
    ensembles : a list of dicts. They contain:
      *  objects like :py:class:`.PathEnsemble`
             These objects contain information about accepted paths for the
             different path ensembles.
    step : integer
        The current step number.
    variables : dict of objects
        This dict contains variables we can use for updating derived data
        for the different path ensembles.

    Returns
    -------
    out : dict of floats
        A dictionary with computed initial flux, crossing probability and
        errors in these.

    """
    calc = {'flux': 0, 'fluxe': 0,
            'pcross': 0, 'pcrosse': float('inf'),
            'kab': 0, 'kabe': float('inf')}
    length0 = variables['length0']
    length1 = variables['length1']
    length0.add(ensembles[0]['path_ensemble'].last_path.length)
    length1.add(ensembles[1]['path_ensemble'].last_path.length)
    flux = 1.0 / ((length0.mean + length1.mean - 4.0) * TIMESTEP)
    calc['flux'] = flux
    variables['flux'].append(flux)
    fluxe = simple_block(variables['flux'], ANALYSIS['nblock'])
    calc['fluxe'] = fluxe
    accprob = 1.0
    matched_prob = []
    matched_lamb = []
    for i, ensemble in enumerate(ensembles):
        if i == 0:
            continue
        prun = variables['prun'][i]
        orderp = variables['orderp'][i]
        prun, ordernew, lamb, pcross = probability_path_ensemble(
            ensemble['path_ensemble'], step, prun, orderp)
        variables['orderp'][i] = ordernew
        variables['prun'][i] = prun
        variables['pcross'][i] = pcross
        variables['lamb'][i] = lamb
        idx = np.where(lamb <= ensemble['path_ensemble'].detect)[0]
        matched_lamb.extend(lamb[idx])
        matched_prob.extend(pcross[idx] * accprob)
        variables['pcross2'][i] = pcross*accprob
        accprob *= prun
    variables['mlamb'] = matched_lamb
    variables['mpcross'] = matched_prob
    if matched_lamb[-1] < INTERFACES[-1]:
        pcross = 0
        pcrosse = float('inf')
    else:
        pcross = matched_prob[-1]
        variables['match'].append(pcross)
        pcrosse = simple_block(variables['match'], ANALYSIS['nblock'])
    calc['pcross'] = pcross
    calc['pcrosse'] = pcrosse
    calc['kab'] = flux * pcross
    if pcross == 0 or flux == 0:
        calc['kabe'] = float('inf')
    else:
        calc['kabe'] = calc['kab'] * np.sqrt(pcrosse**2 / pcross**2 +
                                             fluxe**2 / flux**2)
    return calc


def matplotlib_setup():
    """Set up matplotlib & prepare animation."""
    new_fig = plt.figure(figsize=(12, 6))
    grid = gridspec.GridSpec(3, 4)
    ax1 = new_fig.add_subplot(grid[:, :2])
    ax2 = new_fig.add_subplot(grid[0, 2:])
    ax3 = new_fig.add_subplot(grid[1:, 2:])
    axes = (ax1, ax2, ax3)
    plot_patches = {'paths': [],
                    'prob': [],
                    'prob2': [],
                    'matched': None,
                    'fluxline': None,
                    'txtmove': [],
                    'txtcycle': None,
                    'start': [],
                    'end': []}
    k = 0
    for i, pos in enumerate(INTERFACES):
        ax1.axvline(x=pos, lw=2, ls=':', color='#262626')
        ax3.axvline(x=pos, lw=2, ls=':', color='#262626')
        newline, = ax1.plot([], [], lw=3, ls='-', color=COLORS[i])
        plot_patches['paths'].append(newline)
        if PCROSS_LOG:
            newlinep2, = ax3.plot([], [], lw=3, ls='-',
                                  color=newline.get_color())
            plot_patches['prob2'].append(newlinep2)
        else:
            newlinep, = ax3.plot([], [], lw=3, ls='-',
                                 color=newline.get_color())
            plot_patches['prob'].append(newlinep)

        newscat = ax1.scatter(None, None, s=75, marker='o',
                              color=newline.get_color())
        plot_patches['start'].append(newscat)
        newscat = ax1.scatter(None, None, s=75, marker='s',
                              color=newline.get_color())
        plot_patches['end'].append(newscat)
        # Note, we also add a line for [0^-], this is just to get
        # the colors right, plines[0] is not used for anything else...
        if i == 0:
            txt_x = pos-0.4
            txt_y = 1.8
        else:
            txt_x = INTERFACES[i-1]
            txt_y = 1.8 - (k-1) * 0.2
        txtbox = ax1.text(txt_x, txt_y, '', transform=ax1.transData,
                          backgroundcolor='w', fontsize=12)
        plot_patches['txtmove'].append(txtbox)
        if k % 3 == 0 and k > 0:
            k = 0
        k += 1
    plot_patches['txtcycle'] = ax1.text(-1.4, -1.8, '',
                                        transform=ax1.transData,
                                        backgroundcolor='w', fontsize=14)
    ax1.set_xlabel(r'Order parameter ($\lambda$)')
    ax1.set_ylabel(r'Velocity ($\dot{\lambda}$)')
    if PCROSS_LOG:
        ax3.set_yscale('log')
        plot_patches['matched'] = ax3.plot([], [], lw=6, ls='-',
                                           color='#262626',
                                           zorder=0, alpha=0.7)[0]
        ax3.set_xlim(-1, 1.05)
        ax3.set_ylim(1e-7, 1)
    else:
        axp = ax3.twinx()
        axp.set_yscale('log')
        plot_patches['matched'] = axp.plot([], [], lw=6, ls='-',
                                           color='#262626',
                                           zorder=0, alpha=0.7)[0]
        ax3.set_xlim(-1, 1.05)
        axp.set_ylim(1e-7, 1)
        ax3.set_ylim(0, 1)
    ax3.set_xlabel(r'$\lambda$')
    ax3.set_ylabel(r'Probability')

    ax1.set_ylim(-2, 2)
    ax1.set_xlim(-1.5, 1.5)

    plot_patches['fluxline'] = ax2.plot([0], [0], lw=3, ls='-',
                                        color='#4C72B0')[0]
    ax2.set_ylim(0, 1)
    ax2.set_xlim(0, 1)
    ax2.set_xticklabels([])
    ax2.set_ylabel('Flux')
    ax2.set_xlabel('Cycles completed')
    ax2.locator_params(axis='y', nbins=4)
    if PCROSS_LOG:
        new_fig.subplots_adjust(left=0.1, bottom=0.15, right=0.95, top=0.95,
                                wspace=0.5, hspace=0.5)
    else:
        new_fig.set_tight_layout(True)
    return new_fig, plot_patches, axes


def analyse_prob(ensemble, props, idx, step):
    """Update running estimates for probability.

    Parameters
    ----------
    ensemble : object like :py:class:`.PathEnsemble`
        The ensemble to analyse.
    props : dict
        A dictionary for storing properties we calculate.
    idx : int
        An index for the path ensemble.
    step : int
        The current simulation step.

    """
    orderp = ensemble.last_path.ordermax[0]
    success = 1 if orderp > ensemble.detect else 0
    prun = props['prun'][idx]
    if not prun:
        prun.append(success)
    else:
        npath = step + 1
        prun.append(float(success + prun[-1] * (npath-1)) / float(npath))
    props['orderp'][idx].append(orderp)
    orderparam = np.array(props['orderp'][idx])
    ordermax = min(orderparam.max(), max(ensemble.interfaces))
    ordermin = ensemble.interfaces[1]
    pcross, lamb = _pcross_lambda_cumulative(orderparam, ordermin, ordermax,
                                             ANALYSIS['ngrid'])
    props['pcross'][idx] = (lamb, pcross)


def update(frame, simulation, plot_patches, variables, axes):
    """Run the simulation and update plots.

    Parameters
    ----------
    frame : int
        The current frame number, supplied by animation.FuncAnimation.
    simulation : object like :py:class:`.Simulation`
        The simulation we are running.
    plot_patches : dict of objects
        This dict contains the lines, text boxes etc. from matplotlib
        which we will use to display our data.
    variables : dict of objects
        A dict with results from the simulation.
    axes : tuple
        This tuple contains the axes used for plotting.

    Returns
    -------
    out : list
        list of the patches to be drawn.

    """
    patches = []
    if not simulation.is_finished():
        result = simulation.step()
        step = result['cycle']['step']
        ensembles = simulation.ensembles
        print_to_screen(f'# Current cycle: {step}', level='info')
        anr = analyse_path_ensembles(ensembles, step, variables)
        retis_txt, force = step_txt(ensembles, result, variables['prun'])
        global FTOT
        FTOT += force
        for line in retis_txt:
            print_to_screen(f'# {line}')
        print_to_screen('# Flux: {flux:<8.6g} +- {fluxe:<8.6g}'.format(**anr))
        print_to_screen(('# Crossing probability: {pcross:<8.6g} +-'
                         '{pcrosse:<8.6g}').format(**anr))
        print_to_screen('# K_AB: {kab:<8.6g} +- {kabe:<8.6g}'.format(**anr))
        print_to_screen(f'# No. of force evaluations: {force:g}')
        print_to_screen('')

        for i, ensemble in enumerate(ensembles):
            _, pos, vel = get_path(ensemble['path_ensemble'].last_path)
            plot_patches['paths'][i].set_data(pos, vel)
            patches.append(plot_patches['paths'][i])
            plot_patches['start'][i].set_offsets((pos[0][0], vel[0][0]))
            plot_patches['start'][i].set_visible(True)
            patches.append(plot_patches['start'][i])
            plot_patches['end'][i].set_offsets((pos[-1][0], vel[-1][0]))
            plot_patches['end'][i].set_visible(True)
            patches.append(plot_patches['end'][i])
            if i > 0:
                if not PCROSS_LOG:
                    plot_patches['prob'][i].set_data(variables['lamb'][i],
                                                     variables['pcross'][i])
                    patches.append(plot_patches['prob'][i])

        fluxx, fluxy = plot_patches['fluxline'].get_data()
        fluxx = np.append(fluxx, fluxx[-1] + 1)
        fluxy = np.append(fluxy, anr['flux'])
        plot_patches['fluxline'].set_data(fluxx, fluxy)
        axes[1].set_xlim(0, step+1)
        axes[1].set_ylim(0, 1.1*fluxy.max())
        patches.append(plot_patches['fluxline'])
        for key in result:
            if key.startswith('move'):
                i = int(key.split('-')[1])
                if result[key] == 'tis':
                    trial = result[f'path-{i}']
                    txtmove = trial.generated[0].upper()
                else:
                    txtmove = result[key][:2].upper()
                plot_patches['txtmove'][i].set_text(txtmove)
                plot_patches['txtmove'][i].set_color(TXTCOLOR[txtmove])
                patches.append(plot_patches['txtmove'][i])
        plot_patches['txtcycle'].set_text(f'Cycle: {step}')
        patches.append(plot_patches['txtcycle'])

        # match probabilities:
        for i, ensemble in enumerate(simulation.ensembles):
            if i == 0:
                continue
            if PCROSS_LOG:
                prob2 = variables['pcross2'][i]
                plot_patches['prob2'][i].set_data(variables['lamb'][i], prob2)
                patches.append(plot_patches['prob2'][i])
        plot_patches['matched'].set_data(variables['mlamb'],
                                         variables['mpcross'])
        patches.append(plot_patches['matched'])
        return patches
    # Just return without updating:
    print_to_screen(f'# Simulation is done (frame = {frame})')
    print_to_screen(f'# Total number of force evaluations: {FTOT}')
    patches = []
    for key in ('paths', 'prob', 'prob2', 'txtmove', 'start', 'end'):
        patches.extend(plot_patches[key])
    for key in ('matched', 'fluxline', 'txtcycle'):
        patches.append(plot_patches[key])
    return patches


def main():
    """Run the simulation."""
    colorama.init(autoreset=True)
    print_to_screen('# CREATING SYSTEM', level='info')
    simulation = create_simulation(SETTINGS)
    print_to_screen(simulation)
    print_to_screen('# GENERATING INITIAL PATHS', level='info')

    for i, _ in enumerate(initiate_path_simulation(simulation, SETTINGS)):
        path_ensemble = simulation.ensembles[i]['path_ensemble']
        name = path_ensemble.ensemble_name
        print_to_screen(f'Info about ensemble {name}:',
                        level='success')
        print_to_screen(path_ensemble)
        print_to_screen('Info about the initial path:', level='success')
        print_to_screen(path_ensemble.last_path)
        print_to_screen('')
    # We make a dictionary of these variable for easier access:
    # Set up some variables for storing results:
    fig, plot_patches, axes = matplotlib_setup()
    variables = {'length0': Property('Path length in [0^-]'),
                 'length1': Property('Path length in [0^+]'),
                 'flux': [],
                 'match': [],
                 'orderp': [None for _ in INTERFACES],
                 'prun': [None for _ in INTERFACES],
                 'lamb': [None for _ in INTERFACES],
                 'mlamb': [None for _ in INTERFACES],
                 'pcross': [None for _ in INTERFACES],
                 'mpcross': [None for _ in INTERFACES],
                 'pcross2': [None for _ in INTERFACES]}

    _ = animation.FuncAnimation(fig, update,
                                frames=SETTINGS['simulation']['steps']+1,
                                fargs=[simulation, plot_patches, variables,
                                       axes],
                                repeat=False,
                                interval=10,
                                blit=use_blit())
    plt.show()


def use_blit():
    """Check if we should use blitting."""
    return not sys.platform.startswith('darwin')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module defines some helper functions for pair potentials.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

mixing_parameters (:py:func:`.mixing_parameters`)
    Definition of mixing rules which are used in some force fields
    for generating cross interactions.

generate_pair_interactions (:py:func:`generate_pair_interactions`)
    Function to generate pair parameters from atom parameters.
"""
import itertools
import logging
import numpy as np
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['mixing_parameters', 'generate_pair_interactions']


def _check_pair_parameters(parameters):
    """Check that the required pair parameters are given.

    If the parameters are not given, we will just set them to default
    values.

    Parameters
    ----------
    parameters : dict
        The settings for the potential.

    Note
    ----
    This function will not return anything, but it will update the input
    `parameters` with default values.

    """
    for pair in parameters:
        for key in ('epsilon', 'sigma'):
            if key not in parameters[pair]:
                msg = f'{key} for {pair} not given. Set to 0.0'
                logger.warning(msg)
                parameters[pair][key] = 0.0
        if 'rcut' not in parameters[pair]:
            factor = parameters[pair].get('factor', 0.0)
            rcut = factor * parameters[pair]['sigma']
            parameters[pair]['rcut'] = rcut
            msg = ('rcut for {} not given. Using factor to set it to: '
                   '{} * {} = {}')
            msg = msg.format(pair, parameters[pair]['sigma'], factor, rcut)
            logger.info(msg)


def generate_pair_interactions(parameters, mixing):
    """Generate pair parameters from atom parameters.

    The parameters are given as a dictionary where the keys are
    either just integers -- which defines atom parameters -- or tuples
    which define pair interactions.

    Parameters
    ----------
    parameters : dict
        This dict contain the atom parameters.
    mixing : string
        Determines how we should mix pair interactions.

    """
    _check_pair_parameters(parameters)
    atoms = []
    pair_param = {}
    for key in parameters:
        if key == 'mixing' or isinstance(key, tuple):
            continue
        atoms.append(key)
    for (atmi, atmj) in itertools.product(atoms, atoms):
        pari = parameters[atmi]
        parj = parameters[atmj]
        if (atmi, atmj) in pair_param:
            continue
        if (atmi, atmj) in parameters:
            pair_param[atmi, atmj] = dict(parameters[atmi, atmj])
            pair_param[atmj, atmi] = pair_param[atmi, atmj]
            continue
        if (atmj, atmi) in parameters:
            pair_param[atmj, atmi] = dict(parameters[atmj, atmi])
            pair_param[atmi, atmj] = pair_param[atmj, atmi]
            continue
        if atmi == atmj:
            eps_ij = pari['epsilon']
            sig_ij = pari['sigma']
            rcut_ij = pari['rcut']
            pair_param[atmi, atmi] = {'epsilon': eps_ij, 'sigma': sig_ij,
                                      'rcut': rcut_ij}
        else:
            eps_ij, sig_ij, rcut_ij = mixing_parameters([pari['epsilon'],
                                                         parj['epsilon']],
                                                        [pari['sigma'],
                                                         parj['sigma']],
                                                        [pari['rcut'],
                                                         parj['rcut']],
                                                        mixing)
            pair_param[atmi, atmj] = {'epsilon': eps_ij, 'sigma': sig_ij,
                                      'rcut': rcut_ij}
            pair_param[atmj, atmi] = pair_param[atmi, atmj]
    #     if (atmj, atmi) not in pair_param:
    #         pair_param[atmj, atmi] = pair_param[atmi, atmj]
    return pair_param


def mixing_parameters(epsilon, sigma, rcut, mixing='geometric'):
    r"""Define the so-called mixing rules.

    These mixing rules are used for some force fields when generating
    cross interactions. The known mixing rules are:

    1. Geometric:

       * .. math::

            \epsilon_{ij} = \sqrt{\epsilon_{i} \times \epsilon_{j}}

       * .. math::

            \sigma_{ij} = \sqrt{\sigma_{i} \times \sigma_{j}}

       * .. math::

            r_{\text{c},ij} = \sqrt{r_{\text{c},i} \times r_{\text{c},j}}

    2. Arithmetic:

       * .. math::

            \epsilon_{ij} = \sqrt{\epsilon_{i} \times \epsilon_{j}}

       * .. math::

            \sigma_{ij} = \frac{\sigma_{i} \times \sigma_{j}}{2}

       * .. math::

            r_{\text{c},ij} = \frac{r_{\text{c},i} \times r_{\text{c},j}}{2}

    3. Sixthpower

       * .. math::

            \epsilon_{ij} = 2 \sqrt{\epsilon_{i} \times \epsilon_{j}}
            \frac{\sigma_i^3 \times \sigma_j^3}{\sigma_i^6 + \sigma_j^6}

       * .. math::

            \sigma_{ij} = \left( \frac{\sigma_{i}^6 \times
            \sigma_{j}^6}{2} \right)^{1/6}

       * .. math::

            r_{\text{c},ij} = \left(\frac{r_{\text{c},i}^6 \times
            r_{\text{c},j}^6}{2}\right)^{1/6}


    Parameters
    ----------
    epsilon : list of float
        Lennard-Jones epsilon parameter for a particle of type `i` and 'j'.
    sigma : list of float
        Lennard-Jones sigma parameter for a particle of type `i` and 'j'.
    rcut : list of float
        Lennard-Jones cut-off value for a particle of type `i` and 'j'.
    mixing :  string, optional
        Represents what kind of mixing that should be done.

    Returns
    -------
    out[0] : float
        The mixed ``epsilon_ij`` parameter.
    out[1] : float
        The mixed ``sigma_ij`` parameter.
    out[2] : float
        The mixed ``rcut_ij`` parameter.

    """
    if mixing == 'geometric':
        epsilon_ij = np.sqrt(epsilon[0] * epsilon[1])
        sigma_ij = np.sqrt(sigma[0] * sigma[1])
        rcut_ij = np.sqrt(rcut[0] * rcut[1])
    elif mixing == 'arithmetic':
        epsilon_ij = np.sqrt(epsilon[0] * epsilon[1])
        sigma_ij = 0.5 * (sigma[0] + sigma[1])
        rcut_ij = 0.5 * (rcut[0] + rcut[1])
    elif mixing == 'sixthpower':
        si3 = sigma[0]**3
        si6 = si3**2
        sj3 = sigma[1]**3
        sj6 = sj3**2
        avgs6 = 0.5 * (si6 + sj6)
        epsilon_ij = np.sqrt(epsilon[0] * epsilon[1]) * si3 * sj3 / avgs6
        sigma_ij = avgs6**(1.0 / 6.0)
        rcut_ij = (0.5 * (rcut[0]**6 + rcut[1]**6))**(1.0 / 6.0)
    else:
        epsilon_ij = 0.0
        sigma_ij = 0.0
        rcut_ij = 0.0
        logger.warning('No mixing requested. Cross interactions set to 0')
    return epsilon_ij, sigma_ij, rcut_ij



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the classes and methods from pyretis.core.particles"""
import logging
import numpy as np
import unittest
from pyretis.core.particles import (
    Particles,
    ParticlesExt,
    get_particle_type,
    particles_from_restart,
)
from scipy.special import comb
logging.disable(logging.CRITICAL)


class ParticleTest(unittest.TestCase):
    """Run the tests for the Particle class."""

    def test_creation(self):
        """Test the creation a particle list."""
        for dim in range(1, 4):
            particles = Particles(dim=dim)
            particles.add_particle(np.zeros(dim), np.zeros(dim), np.zeros(dim))
            particles.add_particle(np.zeros(dim), np.zeros(dim), np.zeros(dim))
            self.assertEqual(particles.dim, dim)
            self.assertEqual(particles.npart, 2)
            self.assertIsNone(particles.virial)

    def test_empty_list(self):
        """Test that we can empty the list."""
        particles = Particles(dim=3)
        for _ in range(10):
            particles.add_particle(np.zeros(3), np.zeros(3), np.zeros(3))
        self.assertEqual(particles.npart, 10)
        particles.empty_list()
        self.assertEqual(particles.npart, 0)
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        self.assertEqual(particles.npart, 1)

    def test_pos_setget(self):
        """Test that we can set and get positions."""
        particles = Particles(dim=3)
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        pos = particles.get_pos()
        self.assertFalse(pos is particles.pos)
        pos[0][1] = 100
        for i in particles.pos:
            self.assertTrue(np.allclose(i, np.ones(3)))
        self.assertFalse(np.allclose(pos[0], np.ones(3)))
        particles.set_pos(pos)
        self.assertFalse(pos is particles.pos)
        for i, j in zip(pos, particles.pos):
            self.assertTrue(np.allclose(i, j))

    def test_vel_setget(self):
        """Test that we can set and get velocities."""
        particles = Particles(dim=3)
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        vel = particles.get_vel()
        self.assertFalse(vel is particles.vel)
        vel[0][1] = 100
        for i in particles.vel:
            self.assertTrue(np.allclose(i, np.ones(3)))
        self.assertFalse(np.allclose(vel[0], np.ones(3)))
        particles.set_vel(vel)
        self.assertFalse(vel is particles.vel)
        for i, j in zip(vel, particles.vel):
            self.assertTrue(np.allclose(i, j))

    def test_force_setget(self):
        """Test that we can set and get forces."""
        particles = Particles(dim=3)
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        force = particles.get_force()
        self.assertFalse(force is particles.force)
        force[0][1] = 100
        for i in particles.force:
            self.assertTrue(np.allclose(i, np.ones(3)))
        self.assertFalse(np.allclose(force[0], np.ones(3)))
        particles.set_force(force)
        self.assertFalse(force is particles.force)
        for i, j in zip(force, particles.force):
            self.assertTrue(np.allclose(i, j))

    def test_get_selection(self):
        """Test that we can get a selection of properties."""
        particles = Particles(dim=3)
        for i in range(10):
            particles.add_particle(np.ones(3) * i, np.ones(3) * i,
                                   np.ones(3) * i, name=f'H{i}')
        force = particles.get_selection(['force', 'name'], selection=None)[0]
        self.assertTrue(np.allclose(force, particles.force))
        force[0] = np.ones(3) * 101.
        self.assertTrue(np.allclose(force[0], np.array([101., 101., 101.])))
        out = particles.get_selection(['force', 'name'], selection=[1, 3])
        self.assertTrue(np.allclose(out[0][0], particles.force[1]))
        self.assertTrue(np.allclose(out[0][1], particles.force[3]))
        self.assertEqual(out[1][0], 'H1')
        self.assertEqual(out[1][1], 'H3')

    def test_iterate(self):
        """Test that we can iterate over the particles."""
        particles = Particles(dim=3)
        for i in range(10):
            particles.add_particle(np.ones(3) * i, np.ones(3) * i,
                                   np.ones(3) * i, name=f'A{i}')
        for i, part in enumerate(particles):
            self.assertTrue(np.allclose(part['pos'], particles.pos[i]))

    def test_pairs(self):
        """Test that we can iterate over pairs."""
        particles = Particles(dim=3)
        npart = 21
        for i in range(npart):
            particles.add_particle(np.ones(3) * i, np.ones(3) * i,
                                   np.ones(3) * i, name=f'A{i}')
        pairs = set()
        for pair in particles.pairs():
            pairs.add((pair[0], pair[1]))
        self.assertAlmostEqual(len(pairs), comb(npart, 2))

    def test_generate_restart_info(self):
        """Test that we can generate restart info."""
        particles = Particles(dim=3)
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        particles.add_particle(np.ones(3) * 10, np.ones(3), np.ones(3))
        particles.vpot = 101
        particles.ekin = 102
        restart = particles.restart_info()
        self.assertEqual(restart['type'], 'internal')
        particles2 = Particles(dim=3)
        particles2.load_restart_info(restart)
        self.assertTrue(np.allclose(particles.pos, particles2.pos))

    def test_generate_restart_missing(self):
        """Test that we can generate restart info if attribs. are missing."""
        particles = Particles(dim=3)
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        particles.add_particle(np.ones(3) * 10, np.ones(3), np.ones(3))
        particles.vpot = 101
        del particles.ekin
        restart = particles.restart_info()
        particles2 = Particles(dim=3)
        particles2.load_restart_info(restart)
        self.assertTrue(np.allclose(particles.pos, particles2.pos))
        self.assertTrue(hasattr(particles2, 'ekin'))
        self.assertFalse(hasattr(particles, 'ekin'))
        # Test load and create:
        simulation_restart = {'particles': restart}
        particles3 = particles_from_restart(simulation_restart)
        self.assertNotEqual(particles, particles3)
        particles.ekin = None
        self.assertEqual(particles, particles3)
        particles3 = particles_from_restart({})
        self.assertIsNone(particles3)

    def test_copy(self):
        """Test that we can copy particles."""
        particles = Particles(dim=3)
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        particles.add_particle(np.ones(3), np.ones(3), np.ones(3))
        particles.ekin = 102
        particles2 = particles.copy()
        self.assertIsNot(particles, particles2)
        self.assertEqual(particles, particles2)
        del particles.ekin
        particles2 = particles.copy()
        self.assertIsNot(particles, particles2)
        self.assertNotEqual(particles, particles2)
        # Test that we can change one without changing the other:
        particles2 = particles.copy()
        for attr in ('pos', 'vel', 'force'):
            val1 = getattr(particles2, attr)
            setattr(particles2, attr, np.zeros_like(val1))
            val1 = getattr(particles2, attr)
            val2 = getattr(particles, attr)
            self.assertFalse(np.allclose(val1, val2))


class GetParticleTest(unittest.TestCase):
    """Test that we can get the correct classes."""

    def test_get(self):
        """Test that we get correct classes."""
        cls1 = get_particle_type('internal')
        self.assertTrue(cls1 is Particles)
        cls2 = get_particle_type('external')
        self.assertTrue(cls2 is ParticlesExt)
        with self.assertRaises(ValueError):
            get_particle_type('missing someone')


class ParticleExtTest(unittest.TestCase):
    """Run the tests for the Particle class."""

    def test_creation(self):
        """Test the creation a particle list."""
        particles = ParticlesExt(dim=3)
        self.assertTrue(hasattr(particles, 'pos'))
        self.assertTrue(hasattr(particles, 'vel'))

    def test_empty_list(self):
        """Test that we can empty the list."""
        particles = ParticlesExt(dim=3)
        particles.set_pos(('test', 100))
        self.assertEqual(particles.config, ('test', 100))
        particles.empty_list()
        self.assertEqual(particles.config, (None, None))

    def test_setget_pos(self):
        """Test that we can set/get positions."""
        particles = ParticlesExt(dim=3)
        pos = ('some file', 101)
        self.assertEqual(particles.get_pos(), (None, None))
        particles.set_pos(pos)
        for i, j in zip(particles.get_pos(), pos):
            self.assertEqual(i, j)
        self.assertFalse(pos is particles.get_pos())
        pos2 = particles.get_pos()
        self.assertTrue(pos2 is particles.config)

    def test_set_vel(self):
        """Test that we can set the time direction for positions."""
        particles = ParticlesExt(dim=3)
        particles.set_vel(True)
        self.assertTrue(particles.vel_rev)
        particles.set_vel(False)
        self.assertFalse(particles.vel_rev)

    def test_restart_info(self):
        """Test restart methods."""
        particles = ParticlesExt(dim=3)
        pos = ('skinny love', 2)
        particles.set_pos(pos)
        particles.set_vel(True)
        restart = particles.restart_info()
        self.assertTrue(restart['vel_rev'])
        for i, j in zip(particles.get_pos(), pos):
            self.assertEqual(i, j)
        pos2 = ('My my my', 4)
        restart['config'] = pos2
        particles2 = ParticlesExt(dim=3)
        particles2.load_restart_info(restart)
        for i, j in zip(particles2.get_pos(), pos2):
            self.assertEqual(i, j)
        del restart['config']
        with self.assertRaises(ValueError):
            particles2.load_restart_info(restart)

    def test_copy(self):
        """Test that we can copy particles."""
        particles = ParticlesExt(dim=3)
        particles.set_pos(('filename.ext', 987))
        particles.ekin = 123
        particles.vpot = 456
        particles2 = particles.copy()
        self.assertIsNot(particles, particles2)
        self.assertEqual(particles, particles2)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test methods from pyretis.core.common"""
import os
import sys
import inspect
import logging
import unittest
import numpy as np
from io import StringIO
from pyretis.core.common import (
    big_fat_comparer,
    crossing_finder,
    generic_factory,
    import_from,
    inspect_function,
    numpy_allclose,
    segments_counter,
    select_and_trim_a_segment,
    trim_path_between_interfaces,
    wirefence_weight_and_pick,
    _pick_out_arg_kwargs,
)
from unittest.mock import patch
from .test_path import PATHTEST0, PATHTEST1, PATHTEST2

LOCAL_DIR = os.path.abspath(os.path.dirname(__file__))

logging.disable(logging.CRITICAL)


# Define some functions used for testing:
# pylint: disable=unused-argument
def function1(arg1, arg2, arg3, arg4):
    """To test positional arguments."""
# pylint: enable=unused-argument


# pylint: disable=unused-argument
def function2(arg1, arg2, arg3, arg4=10):
    """To test positional and keyword arguments."""
# pylint: enable=unused-argument


# pylint: disable=unused-argument
def function3(arg1, arg2, arg3, arg4=100, arg5=10):
    """To test positional and keyword arguments."""
# pylint: enable=unused-argument


# pylint: disable=unused-argument
def function4(*args, **kwargs):
    """To test positional and keyword arguments."""
# pylint: enable=unused-argument


# pylint: disable=unused-argument
def function5(arg1, arg2, *args, arg3=100, arg4=100):
    """To test positional and keyword arguments."""
# pylint: enable=unused-argument


# pylint: disable=unused-argument
def function6(arg1, arg2, arg3=100, *, arg4=10):
    """To test positional and keyword arguments."""
# pylint: enable=unused-argument


# pylint: disable=unused-argument, keyword-arg-before-vararg
def function7(arg1, arg2, arg3=100, *args, arg4, arg5=10):
    """To test positional and keyword arguments."""
# pylint: enable=unused-argument, keyword-arg-before-vararg


# pylint: disable=unused-argument
def function8(arg1, arg2=100, self='something'):
    """To test redefinition of __init__."""
# pylint: enable=unused-argument


CORRECT = [
    {'args': ['arg1', 'arg2', 'arg3', 'arg4'],
     'varargs': [], 'kwargs': [], 'keywords': []},
    {'args': ['arg1', 'arg2', 'arg3'],
     'varargs': [], 'kwargs': ['arg4'], 'keywords': []},
    {'args': ['arg1', 'arg2', 'arg3'], 'varargs': [],
     'kwargs': ['arg4', 'arg5'], 'keywords': []},
    {'args': [], 'varargs': ['args'], 'kwargs': [],
     'keywords': ['kwargs']},
    {'args': ['arg1', 'arg2'], 'kwargs': ['arg3', 'arg4'],
     'varargs': ['args'], 'keywords': []},
    {'args': ['arg1', 'arg2'], 'kwargs': ['arg3', 'arg4'],
     'varargs': [], 'keywords': []},
    {'args': ['arg1', 'arg2'], 'kwargs': ['arg3', 'arg4', 'arg5'],
     'varargs': ['args'], 'keywords': []},
]


class TestImportFrom(unittest.TestCase):
    """Test that we can import from other modules."""

    def test_import(self):
        """Test that we can import."""
        module = os.path.join(LOCAL_DIR, 'fooengine.py')
        klass = 'FooEngine'
        imp = import_from(module, klass)
        sys.path.insert(0, LOCAL_DIR)
        from fooengine import FooEngine
        del sys.path[0]
        self.assertEqual(
            inspect.getsource(imp),
            inspect.getsource(FooEngine)
        )

    def test_import_importerror(self):
        """Test that we can import."""
        module = os.path.join(LOCAL_DIR, 'fooengine.py')
        klass = 'DoesNotExist'
        with self.assertRaises(ValueError):
            import_from(module, klass)
        module = 'path-that-should-not-be'
        with self.assertRaises(ValueError):
            import_from(module, klass)
        module = os.path.join(LOCAL_DIR, 'isnothere.py')
        with self.assertRaises(ValueError):
            import_from(module, klass)


class InspectTest(unittest.TestCase):
    """Run the inspect_function method."""

    def test_inspect(self):
        """Test the inspect function method."""
        # Define some test functions:
        functions = [function1, function2, function3, function4,
                     function5, function6, function7]

        for i, func in enumerate(functions):
            args = inspect_function(func)
            self.assertEqual(args, CORRECT[i])

    def test_arg_kind(self):
        """Test a function with only positional arguments."""
        args = inspect_function(range.__eq__)
        self.assertTrue(not args['keywords'])
        self.assertTrue(not args['varargs'])
        self.assertTrue(not args['kwargs'])
        for i in ('self', 'value'):
            self.assertTrue(i in args['args'])

    def test_pick_out_kwargs(self):
        """Test pick out of "self" for kwargs."""
        settings = {'arg1': 10, 'arg2': 100, 'self': 'text'}

        # pylint: disable=too-few-public-methods
        class Abomination:
            """Just to allow redefination of __init__."""
        # pylint: enable=too-few-public-methods

        abo = Abomination()
        abo.__init__ = function8

        args, kwargs = _pick_out_arg_kwargs(abo, settings)
        self.assertFalse('self' in args)
        self.assertFalse('self' in kwargs)

        with self.assertRaises(ValueError):
            _, _ = _pick_out_arg_kwargs(abo, {})


class Klass1:
    """A class used for factory testing."""

    def __init__(self):
        self.stuff = 10

    def method1(self):
        """Return stuff."""
        return self.stuff

    def method2(self, add):
        """Add to stuff."""
        self.stuff += add


class TestGenericFactory(unittest.TestCase):
    """Test the generic factory."""

    def test_factory(self):
        """Test that the factory works as intended."""
        factory_map = {'klass1': {'cls': Klass1}}

        settings = {'class': 'Klass1'}
        cls = generic_factory(settings, factory_map, name='Testing')
        self.assertIsInstance(cls, Klass1)

        settings = {'clAsS': 'Klass1'}
        logging.disable(logging.INFO)
        with self.assertLogs('pyretis.core.common', level='CRITICAL'):
            cls = generic_factory(settings, factory_map, name='Testing')
        logging.disable(logging.CRITICAL)
        self.assertTrue(cls is None)

        settings = {'Klass': 'Klass1'}
        logging.disable(logging.INFO)
        with self.assertLogs('pyretis.core.common', level='CRITICAL'):
            cls = generic_factory(settings, factory_map, name='Testing')
        logging.disable(logging.CRITICAL)
        self.assertTrue(cls is None)


class TestNumpyComparison(unittest.TestCase):
    """Test the numpy_allclose method."""

    def test_numpy_allclose(self):
        """Test that the numpy allclose comparison works as intended."""
        val1 = np.array([1.2345, 2.34567, 9.87654321])
        val2 = np.array([1.2345, 2.34567, 9.87654321])
        val3 = np.array([1.2345, 2.34567, 10.])
        # Both are valid:
        self.assertTrue(numpy_allclose(val1, val2))
        self.assertFalse(numpy_allclose(val1, val3))
        # Both are None:
        self.assertTrue(numpy_allclose(None, None))
        # One is None:
        self.assertFalse(numpy_allclose(val1, None))
        self.assertFalse(numpy_allclose(None, val1))
        # One is something else:
        self.assertFalse(numpy_allclose(val1, 'the roof is on fire'))
        # Both is something else:
        self.assertFalse(numpy_allclose('the roof is on fire',
                                        'the roof is on fire'))


class TestBigFatComparer(unittest.TestCase):
    """Test the big_fat_comparer."""

    def test_fat_part_1(self):
        """Test the big_fat_comparer."""
        self.assertFalse(big_fat_comparer(np.array([1]), np.array([1, 2, 3])))
        with self.assertRaises(ValueError):
            big_fat_comparer(np.array([1]), np.array([1, 2, 3]), hard=True)

        self.assertFalse(big_fat_comparer(np.array([1, 2, 4]),
                                          np.array([1, 2, 3])))
        with self.assertRaises(ValueError):
            big_fat_comparer(np.array([1, 2, 4]),
                             np.array([1, 2, 3]), hard=True)

        self.assertFalse(big_fat_comparer(1, 2))
        with self.assertRaises(ValueError):
            big_fat_comparer(1, 2, hard=True)

        self.assertFalse(big_fat_comparer(1, 2))
        with self.assertRaises(ValueError):
            big_fat_comparer(1, 2, hard=True)

        self.assertFalse(big_fat_comparer(1, [1]))
        with self.assertRaises(ValueError):
            big_fat_comparer(1, [1], hard=True)

        self.assertFalse(big_fat_comparer((1, 3, 4), (1, 3)))
        with self.assertRaises(ValueError):
            big_fat_comparer((1, 3, 4), (1, 3), hard=True)
        self.assertFalse(big_fat_comparer((1, 3), (1, 3, 4)))
        with self.assertRaises(ValueError):
            big_fat_comparer((1, 3), (1, 3, 4), hard=True)
        self.assertFalse(big_fat_comparer((1, 2, 3), (2, 3, 4)))
        with self.assertRaises(ValueError):
            big_fat_comparer((1, 2, 3), (2, 3, 4), hard=True)

        self.assertFalse(big_fat_comparer('Ciao', 'Mamma'))
        with self.assertRaises(ValueError):
            big_fat_comparer('Ciao', 'Mamma', hard=True)

        self.assertFalse(big_fat_comparer([1, 2], [0]))
        with self.assertRaises(ValueError):
            big_fat_comparer([1, 2], [0], hard=True)

        a, b = np.array([0, 2]), np.array([0, 1])
        with self.assertRaises(ValueError):
            big_fat_comparer(a, b, hard=True)

        a, b = [0, 2], [0, {'b': 'c'}]
        with self.assertRaises(ValueError):
            big_fat_comparer(a, b, hard=True)

    def test_fat_part_2(self):
        """Test the big_fat_comparer."""
        try1 = {'correct': CORRECT,
                'kkk': [1, 2, 3],
                'giorgio': {'santo': 'subito',
                            'yes': [12, 'b']}}

        try2 = {'correct': CORRECT,
                'kkk': [1, 2, 3],
                'giorgio': {'santo': 'subito',
                            'yes': [12, 'b']}}

        try3 = {'correct': CORRECT,
                'kkk': [1, 2, 3],
                'giorgio': {'santo': 'subit',
                            'yes': [12, 'b']}}
        try4 = {'correct': CORRECT,
                'kkk': [1, 2, 3],
                'giorgio': {'santo': 'subit',
                            'no': 'interest',
                            'yes': {'b': 12}}}

        self.assertFalse(big_fat_comparer(try1, try3))
        with self.assertRaises(ValueError):
            big_fat_comparer(try1, try3, hard=True)

        self.assertFalse(big_fat_comparer(try1, try4))
        with self.assertRaises(ValueError):
            big_fat_comparer(try1, try4, hard=True)

        self.assertFalse(big_fat_comparer(try3, try4))
        with self.assertRaises(ValueError):
            big_fat_comparer(try3, try4, hard=True)

        self.assertFalse(big_fat_comparer(try2, try4))
        with self.assertRaises(ValueError):
            big_fat_comparer(try2, try4, hard=True)

        self.assertFalse(big_fat_comparer(try1, {**{'Ihate': 'Cinnamon'},
                                                 **try2}))
        with self.assertRaises(ValueError):
            big_fat_comparer(try1, {**{'Ihate': 'Cinnamon'},
                                    **try2}, hard=True)

        self.assertFalse(big_fat_comparer({**{'Iate': 'Cinnamon'}, **try2},
                                          try1))
        with self.assertRaises(ValueError):
            big_fat_comparer({**{'Iate': 'Cinnamon'}, **try2},
                             try1, hard=True)

        self.assertFalse(big_fat_comparer(try1, 'TestMcTest'))
        with self.assertRaises(ValueError):
            big_fat_comparer(try1, 'TestMcTest', hard=True)

        self.assertFalse(big_fat_comparer('TestMcTest', try1))
        with self.assertRaises(ValueError):
            big_fat_comparer('TestMcTest', try1, hard=True)

        self.assertTrue(big_fat_comparer(try1, try2))
        self.assertTrue(big_fat_comparer(try1, try2, hard=True))

        try2['giorgio']['yes'][0] = 1
        self.assertFalse(big_fat_comparer(try1, try2))
        with self.assertRaises(ValueError):
            big_fat_comparer(try1, try2, hard=True)

        try2['giorgio']['yes'][0] = 'c'
        self.assertFalse(big_fat_comparer(try1, try2))
        with self.assertRaises(ValueError):
            big_fat_comparer(try1, try2, hard=True)

        try2['giorgio']['yes'][0] = 12
        try1['giorgio']['yes'].append(12)
        self.assertFalse(big_fat_comparer(try1, try2))
        with self.assertRaises(ValueError):
            big_fat_comparer(try1, try2, hard=True)

        del try1['giorgio']['yes'][2]
        self.assertTrue(big_fat_comparer(try1, try2))
        self.assertTrue(big_fat_comparer(try1, try2, hard=True))


class Test_wt(unittest.TestCase):
    """Run the tests for the wt utilities."""

    def test_x1_crossing_finder(self):
        ph1, ph2 = crossing_finder(path=PATHTEST0, interface=2.9)
        self.assertEqual(15, ph1.order[0])

    def test_segments_counter(self):
        self.assertEqual(4, segments_counter(path=PATHTEST0,
                                             interface_l=2.9, interface_r=4.1))
        self.assertEqual(4, segments_counter(path=PATHTEST0,
                                             interface_l=2, interface_r=9))

    def test_trim_path_between_interfaces(self):
        """Test for trimming trajectories withing a range in path space."""
        path0 = trim_path_between_interfaces(path=PATHTEST0, interface_l=3.0,
                                             interface_r=7.0)
        path2 = PATHTEST2.copy()
        path2.generated = 'ct'

        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path0.length == 12)
            self.assertTrue(path2 == path0)
            self.assertTrue(PATHTEST0 != path0)

    def test_wirefence_weight_and_pick(self):
        """Test for wirefernce function in a pathological case."""
        path0 = trim_path_between_interfaces(path=PATHTEST0, interface_l=3.0,
                                             interface_r=7.0)
        # Create a path with L-L, L-R and R-R sub-paths between 1.5 and 2.5:
        for i, j in enumerate([1]+[2]*3+[1]+[2]*1+[3]+[2]*4+[3]):
            path0.phasepoints[i].order[0] = j

        q_tot, segment = wirefence_weight_and_pick(path=path0,
                                                   intf_l=1.5, intf_r=2.5)
        # Only the L-L and L-R phase points to be counted
        self.assertEqual(q_tot, 4)

        # Create a path with phase points i < intf_l and i+1 > intf_r,
        # and also a R-L sub-path:
        for i, j in enumerate([1]+[3]*6+[2]*4+[1]):
            path0.phasepoints[i].order[0] = j
        q_tot, segment = wirefence_weight_and_pick(path=path0,
                                                   intf_l=1.5, intf_r=2.5,)
        # Only the R-L phase points to be counted
        self.assertEqual(q_tot, 4)

    def test_select_and_trim_a_segment(self):
        """Test to trim a selected segment keeping the crossing."""
        path_rnd = select_and_trim_a_segment(path=PATHTEST0, interface_l=4.1,
                                             interface_r=5.9)
        path3 = select_and_trim_a_segment(path=PATHTEST0, interface_l=4.1,
                                          interface_r=5.9, segment_to_pick=0)
        path1 = PATHTEST1.copy()
        path1.generated = 'sg'

        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path3 != path_rnd)
            self.assertTrue(path1 == path3)
            self.assertTrue(path3.length == 3)
            self.assertTrue(PATHTEST1 != path3)

        path2 = path1.empty_path()
        path2.append(path1.phasepoints[0])
        path2.append(path1.phasepoints[1])

        path3 = select_and_trim_a_segment(path=path2, interface_l=4.1,
                                          interface_r=5.9, segment_to_pick=0)
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path3 != path_rnd)
            self.assertTrue(path3.length == 2)
            self.assertTrue(PATHTEST1 != path3)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the LAMMPSEngine class."""
import logging
import os
import shutil
import tempfile
import unittest
import numpy as np
from pyretis.core.random_gen import create_random_generator
from pyretis.core.system import System
from pyretis.core.particles import ParticlesExt
from pyretis.core.path import Path
from pyretis.engines.lammps import (
    create_lammps_md_input,
    read_lammps_input,
    read_lammps_log,
    system_to_lammps,
    LAMMPSEngine,
)


logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))
INPUT_PATH = os.path.join(HERE, 'lammps_input')
# Define some error messages we expect:
MISSING_FILE_MSG = 'Missing input file'
MISSING_REVVEL_MSG = "'reverse_velocities'"
AIMLESS_ERROR_MSG = 'LAMMPS only support the aimless velocity modification.'


LAMMPS_IN = [
    ('units', 'real'),
    ('timestep', '1.0'),
    ('atom_style', 'full'),
    ('bond_style', 'hybrid morse harmonic'),
    ('angle_style', 'hybrid class2 harmonic'),
    ('dihedral_style', 'class2'),
    ('improper_style', 'class2'),
    ('read_data', 'system.data'),
    ('include', '"groups.in"'),
    ('include', '"md.in.settings"'),
    ('thermo', '100'),
    ('variable', 'SET_TEMP equal 343'),
    ('include', '"order.in"'),
    ('fix', '1 mg nvt temp 343 343 100'),
    ('fix', '2 dummy nvt temp 343 343 10'),
    ('fix', '3 water rigid/nvt molecule temp 343 343 100'),
]


LAMMPS_ENERGY_DATA = {
    'Step': [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
    'Temp': [354.83622, 354.87936, 353.55139, 346.73993, 355.86763,
             356.02159, 351.24368, 338.50584, 350.65277, 341.57213,
             337.38405],
    'Press': [1503.3635, -1496.9831, -2155.9235, 2692.286,
              3461.4022, 522.80391, -190.30433, 55.010487,
              138.91587, 1222.3388, 343.12519],
    'PotEng': [-5047.0936, -4349.3609, -4642.5397, -1634.8362,
               -1475.5791, -3255.0873, -3918.0104, -3858.1462,
               -3498.4035, -3128.2873, -3484.1554],
    'KinEng': [1080.9687, 1081.1002, 1077.0547, 1056.3043,
               1084.1108, 1084.5799, 1070.0245, 1031.2201,
               1068.2243, 1040.5612, 1027.8027],
    'TotEng': [-3966.1248, -3268.2607, -3565.4851, -578.53193,
               -391.46827, -2170.5074, -2847.9859, -2826.9261,
               -2430.1791, -2087.7261, -2456.3527],
}


def get_setting(keyword, settings):
    """Return settings for a given keyword.

    Parameters
    ----------
    keyword : string
        The keyword to find settings for.
    settings : list of tuples
        The settings to search for the keyword in. It is assumed
        that the first item in the tuple is the keyword and the
        second the setting.

    Returns
    -------
    out : list of strings
        The settings found for a given keyword.

    """
    found = []
    for (key, val) in settings:
        if key.lower() == keyword.lower():
            found.append(val)
    return found


def create_system_for_lammps():
    """Create a test system for LAMMPS."""
    system = System()
    system.particles = ParticlesExt(dim=3)
    system.particles.set_pos(('system.data', 0))
    system.particles.set_vel(False)
    return system


def read_seed(lammps_input):
    """Read the seed from the given LAMMPS input file."""
    settings = read_lammps_input(lammps_input)
    seed = None
    for var in settings:
        if var[1].startswith('VEL_SEED'):
            seed = int(var[1].split('equal')[-1])
    return seed


def read_raw_file(filename):
    """Return the lines from a file."""
    lines = []
    with open(filename, 'r', encoding="utf8") as infile:
        lines = [i.strip() for i in infile]
    return lines


class LAMMPSEngineMethodsTest(unittest.TestCase):
    """Run the tests for the methods defined in LAMMPS module."""

    def test_read_lammps_input(self):
        """Test that we can read LAMMPS input."""
        infile = os.path.join(HERE, 'lammps_input', 'lammps.in')
        settings = read_lammps_input(infile)
        self.assertEqual(len(LAMMPS_IN), len(settings))
        for i, j in zip(LAMMPS_IN, settings):
            self.assertEqual(i, j)
        timestep = get_setting('timestep', settings)
        self.assertEqual(timestep, ['1.0'])
        fix = get_setting('fix', settings)
        correct_fix = [item[1] for item in settings[-3:]]
        self.assertTrue(len(correct_fix), len(fix))
        for i, j in zip(correct_fix, fix):
            self.assertEqual(i, j)

    def test_create_md_input(self):
        """Test that we can create input files for LAMMPS."""
        infile = os.path.join(HERE, 'lammps_input', 'lammps.in')
        system = create_system_for_lammps()
        settings = {
            'subcycles': 100,
            'traj': 'test.lammpstrj',
            'steps_subcycles': 12345,
            'timestep': 0.006789,
            'generate_vel': {'seed': 9876, 'rescale': 1},
            'interfaces': [-1, 0, 1],
            'dimension': 3,
            'reverse_velocities': True,
        }
        with tempfile.NamedTemporaryFile() as temp:
            written = create_lammps_md_input(
                system, infile, temp.name, settings
            )
            temp.flush()
            data = []
            with open(temp.name, 'r', encoding="utf8") as infile:
                data = [i.strip() for i in infile]
            # Here written and data will differ, since data contains
            # the full output. But written should be contained in data,
            # check that the last lines are equal:
            offset = len(written)
            for i, line in enumerate(written):
                self.assertEqual(line, data[-(offset-i)])

    def test_read_lammps_log(self):
        """Test that we can read some data from a LAMMPS log file."""
        infile = os.path.join(HERE, 'lammps_input', 'pyretis_md.log')
        logdata = read_lammps_log(infile)
        self.assertIn('energy', logdata)
        self.assertEqual(len(logdata['energy']), len(LAMMPS_ENERGY_DATA))
        for i in LAMMPS_ENERGY_DATA:
            self.assertIn(i, logdata['energy'])
        for key, val in LAMMPS_ENERGY_DATA.items():
            self.assertTrue(np.allclose(val, logdata['energy'][key]))

    def test_system_to_lammps(self):
        """Test the conversion of systems to LAMMPS commands."""
        system = System()
        system.particles = ParticlesExt(dim=3)

        # 1) Test for a restart file:
        system.particles.set_pos(('some.restart', 0))
        system.particles.set_vel(False)
        correct = [
            '# PyRETIS requested a LAMMPS restart file:',
            'read_restart some.restart',
        ]
        cmd = system_to_lammps(system, False, 3)
        self.assertEqual(correct, cmd)

        # 2) Test for a restart file & reversing velocities:
        correct = [
            '# PyRETIS requested a LAMMPS restart file:',
            'read_restart some.restart',
            '# PyRETIS requested reversing of velocities:',
            'variable        vx atom -vx',
            'variable        vy atom -vy',
            'variable        vz atom -vz',
            'velocity        all set v_vx v_vy v_vz',
            'run 0',
        ]
        system.particles.set_pos(('some.restart', 0))
        system.particles.set_vel(True)
        cmd = system_to_lammps(system, True, 3)
        self.assertEqual(correct, cmd)

        # 3) Test if we give a file with an index:
        correct = [
            '# PyRETIS requested the following snapshot:',
            'read_dump notsome.restart 10 x y z vx vy vz ix iy iz box yes',
            '# Reset time step:',
            'reset_timestep 0',
        ]
        system.particles.set_pos(('notsome.restart', 10))
        system.particles.set_vel(False)
        cmd = system_to_lammps(system, False, 3)
        self.assertEqual(correct, cmd)

        # 4) Test index + reversing velocities:
        system.particles.set_pos(('dump.lammpstrj', 10))
        correct = [
            '# PyRETIS requested the following snapshot:',
            'read_dump dump.lammpstrj 10 x y z vx vy vz ix iy iz box yes',
            '# Reset time step:',
            'reset_timestep 0',
            '# PyRETIS requested reversing of velocities:',
            'variable        vx atom -vx',
            'variable        vy atom -vy',
            'variable        vz atom -vz',
            'velocity        all set v_vx v_vy v_vz',
            'run 0',
        ]
        system.particles.set_vel(True)
        cmd = system_to_lammps(system, True, 3)
        self.assertEqual(correct, cmd)


class LAMMPSEngineTest(unittest.TestCase):
    """Run the tests for the LAMMPSEngine class."""

    def test_lammps_initiate(self):
        """Test that we can initiate the LAMMPS engine."""
        # This should not fail:
        engine = LAMMPSEngine('lmp_serial', INPUT_PATH, 10, extra_files=None)
        self.assertEqual(engine.lmp, 'lmp_serial')
        self.assertEqual(engine.engine_type, 'external')
        self.assertEqual(engine.description, 'LAMMPS Engine')
        # This should also fail - input files not found:
        extra_files = ['Oh Lawd He Comin']
        with self.assertRaises(ValueError) as context:
            LAMMPSEngine('lmp_serial', INPUT_PATH, 10,
                         extra_files=extra_files)
        except_str = str(context.exception)
        self.assertTrue(
            except_str.startswith(MISSING_FILE_MSG)
        )
        self.assertTrue(extra_files[0] in except_str)

    def test_run_lammps(self):
        """Test the method to run LAMMPS."""
        # Note: More of the functionality of this method
        # is implicitly tested in the methods that test
        # integration and propagation.
        system = create_system_for_lammps()
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copy(os.path.join(HERE, 'aprogram.py'), tempdir)
            engine = LAMMPSEngine('./aprogram.py', INPUT_PATH,
                                  10, extra_files=None)
            engine.exe_dir = tempdir
            # First we test with missing settings:
            settings = {}
            # None of the required settings were given, engine should fail:
            with self.assertRaises(KeyError) as context:
                engine.run_lammps(system, settings, 'lammps_test')
            except_str = str(context.exception)
            self.assertEqual(except_str, MISSING_REVVEL_MSG)
            # Add the missing required setting(s):
            settings['steps_subcycles'] = 1
            settings['reverse_velocities'] = False
            # By construction of "aprogram.py" this just gives a
            # RuntimeError:
            with self.assertRaises(RuntimeError):
                engine.run_lammps(system, settings, 'lammps_test')
            # Check the standard out and error:
            correct = [
                ['This is a program for testing external commands.'],
                ['ERROR: Program got arguments:',
                 ('./aprogram.py -in lammps_test.in '
                  '-l lammps_test.log -screen '
                  'lammps_test.screen')]
            ]
            for filei, correcti in zip(('stdout.txt', 'stderr.txt'), correct):
                with open(os.path.join(tempdir, filei), 'r',
                          encoding="utf8") as output:
                    lines = output.readlines()
                    self.assertEqual(len(lines), len(correcti))
                    for i, j in zip(lines, correcti):
                        self.assertEqual(i.strip(), j)

    def test_add_files(self):
        """Test that the LAMMPS engine adds input files."""
        extra_files = ['order.in']
        engine = LAMMPSEngine('lammps', INPUT_PATH, 10,
                              extra_files=extra_files)
        with tempfile.TemporaryDirectory() as tempdir:
            engine.add_input_files(tempdir)
            files = [i.name for i in os.scandir(tempdir) if i.is_file()]
            expected = ['system.data'] + extra_files
            self.assertEqual(len(files), len(expected))
            for i in files:
                self.assertIn(i, expected)

    def test_integrate(self):
        """Test the integrate method.

        Note that we do not run LAMMPS here, we just test that we
        can create files etc. for LAMMPS.

        """
        # We expect that the following files will be created:
        expected_files = [
            'pyretis_md.screen',
            'pyretis_md.log',
            'pyretis_md.lammpstrj',
            'order_pyretis_md.txt',
        ]
        steps = 10
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copy(os.path.join(HERE, 'mocklammps.py'), tempdir)
            engine = LAMMPSEngine('./mocklammps.py', INPUT_PATH,
                                  10, extra_files=None)
            engine.exe_dir = tempdir
            engine.add_input_files(tempdir)
            system = create_system_for_lammps()
            # The expected files should not be present yet:
            for i in expected_files:
                self.assertFalse(os.path.isfile(os.path.join(tempdir, i)))
            engine.integrate(system, steps)
            # We now expect to find the files:
            for i in expected_files:
                self.assertTrue(os.path.isfile(os.path.join(tempdir, i)))
            # And we expect the system to have an order parameter:
            self.assertEqual(
                os.path.join(tempdir, 'pyretis_md.lammpstrj'),
                system.particles.get_pos()[0],
            )
            self.assertEqual(
                engine.subcycles * steps, system.particles.get_pos()[1]
            )
            self.assertEqual(len(system.order), 2)
            self.assertEqual(system.order[0], engine.subcycles * steps + 1)
            self.assertEqual(system.order[1], -(engine.subcycles * steps + 1))

    def test_modify_velocities(self):
        """Test the modify_velocities method."""
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copy(os.path.join(HERE, 'mocklammps.py'), tempdir)
            engine = LAMMPSEngine('./mocklammps.py', INPUT_PATH,
                                  10, extra_files=None)
            engine.exe_dir = tempdir
            engine.add_input_files(tempdir)
            system = create_system_for_lammps()
            # Add some fake data:
            system.order = [-1, -1]
            system.particles.ekin = None
            ensemble = {'system': system, 'rgen': None}
            vel_settings = {'aimless': True}
            dek, kin_new = engine.modify_velocities(ensemble, vel_settings)
            self.assertAlmostEqual(4.0, kin_new)
            self.assertEqual(float('inf'), dek)
            self.assertTrue(np.allclose([1., -1.], system.order))
            # Check the calculate order:
            self.assertIs(
                system.order,
                engine.calculate_order(ensemble),
            )
            # Check the created input file for LAMMPS:
            seed = read_seed(os.path.join(tempdir, 'generate_vel.in'))
            self.assertEqual(seed, 1)
            # Test that we can set the seed:
            rgen = create_random_generator({'rgen': 'mock'})
            ensemble = {'system': system, 'rgen': rgen}
            vel_settings = {'aimless': True}
            dek, kin_new = engine.modify_velocities(ensemble, vel_settings)
            self.assertAlmostEqual(4.0, kin_new)
            self.assertAlmostEqual(0.0, dek)
            self.assertTrue(np.allclose([1., -1.], system.order))
            seed = read_seed(os.path.join(tempdir, 'generate_vel.in'))
            self.assertAlmostEqual(1675209429, seed)
            with self.assertRaises(ValueError) as context:
                vel_settings = {'aimless': False}
                engine.modify_velocities(ensemble, vel_settings)
            except_str = str(context.exception)
            self.assertEqual(except_str, AIMLESS_ERROR_MSG)

    def test_propagate_forward(self):
        """Test the propagate method."""
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copy(os.path.join(HERE, 'mocklammps.py'), tempdir)
            engine = LAMMPSEngine('./mocklammps.py', INPUT_PATH,
                                  10, extra_files=None)
            engine.exe_dir = tempdir
            engine.add_input_files(tempdir)
            system = create_system_for_lammps()
            # Run "forward" until we cross the last interface:
            path = Path(rgen=None, maxlen=10)
            ensemble = {
                'system': system,
                'order_function': None,
                'interfaces': [-1.0, 0.0, 80.0]
            }
            success, _ = engine.propagate(
                path=path, ensemble=ensemble, reverse=False
            )
            # Check that the stopping condition is correctly written
            # to the input file.
            correct = (
                'fix op_stop_right all halt 10 v_op_1 > 80.0',
                'fix op_stop_left all halt 10 v_op_1 < -1.0'
            )
            pref = [i for i in os.listdir(tempdir) if i[0].isdigit()][0][0]
            lmp_in = read_raw_file(os.path.join(tempdir, pref + '_trajF.in'))
            for i in correct:
                self.assertIn(i, lmp_in)
            self.assertTrue(success)
            self.assertEqual(path.length, 9)
            traj = [i.particles.get_pos() for i in path.phasepoints]
            for i, point in enumerate(traj):
                self.assertEqual(point[1], i*engine.subcycles)
                self.assertTrue(tempdir in point[0])
                self.assertTrue('trajF.lammpstrj' in point[0])
            self.assertFalse(
                any(i.particles.get_vel() for i in path.phasepoints)
            )
            # Run "forward" until we exceed the max length:
            path = Path(rgen=None, maxlen=4)
            ensemble = {
                'system': system,
                'order_function': None,
                'interfaces': [-1., 0., float('inf')],
            }
            success, _ = engine.propagate(
                path=path, ensemble=ensemble, reverse=False,
            )
            self.assertFalse(success)
            self.assertEqual(path.length, path.maxlen)
            traj = [i.particles.get_pos() for i in path.phasepoints]
            for i, point in enumerate(traj):
                self.assertEqual(point[1], i*engine.subcycles)
                self.assertTrue(tempdir in point[0])
                self.assertTrue('trajF.lammpstrj' in point[0])
            self.assertFalse(
                any(i.particles.get_vel() for i in path.phasepoints)
            )

    def test_propagate_backward(self):
        """Test the propagate method."""
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copy(os.path.join(HERE, 'mocklammps.py'), tempdir)
            engine = LAMMPSEngine('./mocklammps.py', INPUT_PATH,
                                  10, extra_files=None)
            engine.exe_dir = tempdir
            engine.add_input_files(tempdir)
            system = create_system_for_lammps()
            # Run "forward" until we cross the last interface:
            path = Path(rgen=None, maxlen=4)
            ensemble = {
                'system': system,
                'order_function': None,
                'interfaces': [-1, 0.0, 42.0]
            }
            success, _ = engine.propagate(
                path=path, ensemble=ensemble, reverse=True
            )
            self.assertFalse(success)
            self.assertEqual(path.length, 4)
            traj = [i.particles.get_pos() for i in path.phasepoints]
            for i, point in enumerate(traj):
                self.assertEqual(point[1], i*engine.subcycles)
                self.assertTrue('trajB.lammpstrj' in point[0])
                self.assertTrue(tempdir in point[0])
            self.assertTrue(
                all(i.particles.get_vel() for i in path.phasepoints)
            )

    def test_dump_phasepoint(self):
        """Test the method to dump a phase point."""
        system = create_system_for_lammps()
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copy(os.path.join(HERE, 'mocklammps.py'), tempdir)
            engine = LAMMPSEngine('./mocklammps.py', INPUT_PATH,
                                  10, extra_files=None)
            engine.exe_dir = tempdir
            engine.dump_phasepoint(system, deffnm='dumped')
            # Check that the dumped file exists:
            self.assertTrue(
                os.path.isfile(os.path.join(tempdir, 'dumped.lammpstrj'))
            )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test common methods that are used in testing."""
import os
import pathlib
import tempfile
import unittest
from pyretis.inout.common import make_dirs
from pyretis.testing.helpers import search_for_files, clean_dir
from pyretis.testing.systemhelp import create_system_ext


class TestSearchForFiles(unittest.TestCase):
    """Test that we can search for files."""

    def test_search(self):
        """Test the search for files method."""
        dirs = ['these', 'are', 'directories']
        files = ['test1.txt', 'test2.txt', 'test3.txt']
        created_files = []
        created_files_match = []
        with tempfile.TemporaryDirectory() as tempdir:
            # Make some files in the root directory:
            for i in files:
                filename = pathlib.Path().joinpath(tempdir, i)
                pathlib.Path(filename).touch()
                created_files.append(filename)
            # Make some folders with files:
            for i, j in zip(dirs, files):
                dirname = pathlib.Path().joinpath(tempdir, i)
                make_dirs(dirname)
                filename = pathlib.Path().joinpath(dirname, j)
                pathlib.Path(filename).touch()
                created_files.append(filename)
            # Test searching for all files:
            found_files = search_for_files(tempdir)
            self.assertEqual(
                sorted(found_files),
                sorted([str(i) for i in created_files])
            )
            # Test searching for a specific file name:
            found_files = search_for_files(tempdir, match='test4.txt')
            self.assertEqual(len(found_files), 0)
            found_files = search_for_files(tempdir, match='test3.txt')
            created_files_match = [
                str(i) for i in created_files if i.name == 'test3.txt'
            ]
            self.assertEqual(sorted(found_files), sorted(created_files_match))

    def test_clean_dir(self):
        """Test the search for files method."""
        created_files = []
        with tempfile.TemporaryDirectory() as tempdir:
            # Make some files in the root directory:
            for i in range(11):
                filename = pathlib.Path().joinpath(tempdir, f'file{i}')
                pathlib.Path(filename).touch()
                created_files.append(filename)
            files = [i for i in os.scandir(tempdir) if i.is_file]
            self.assertEqual(len(files), len(created_files))
            # Remove the files:
            clean_dir(tempdir)
            files = [i for i in os.scandir(tempdir) if i.is_file]
            self.assertEqual(len(files), 0)


class TestSystemHelp(unittest.TestCase):
    """Test the System helpers."""

    def test_create(self):
        """Test that we can create systems."""
        cases = [
            (None, False),
            (None, True),
            (('file.txt', 1), False),
            (('file.txt', 2), True),
        ]
        for case in cases:
            system = create_system_ext(pos=case[0], vel=case[1])
            if case[0] is None:
                self.assertEqual(system.particles.get_pos(), (None, None))
            else:
                self.assertEqual(system.particles.get_pos(), case[0])
            self.assertEqual(system.particles.get_vel(), case[1])


if __name__ == '__main__':
    unittest.main()
